# -*- coding: utf-8 -*-

"""
Builders for global functions to operate over given ClassRegistry instance.
"""


def register_function(registry_instance):

    def register(cls):
        """Registers new class in the registry."""
        registry_instance.register(cls)

    return register


def unregister_function(registry_instance):

    def unregister(cls):
        """Unegisters the class from the registry."""
        registry_instance.unregister(cls)

    return unregister


def as_choices_function(registry_instance):

    def as_choices():
        """
        Retrieves choice list suitable for form's ChoiceField
        from the registry.
        """
        return registry_instance.as_choices()

    return as_choices


def get_function(registry_instance):

    def get(key):
        """Retrieves class for given key from the registry"""
        return registry_instance.get(key)

    return get



# -*- coding: utf-8 -*-

from django.utils.safestring import mark_safe
from django.contrib import admin

from powerpages.forms import PageAdminForm
from powerpages.models import Page
from powerpages.sync import PageFileDumper, SyncStatus
from powerpages.signals import page_edited


def website_link(page):
    """Link to the Page on website"""
    if not page:
        return None
    url = page.url
    url_parts = url.split('/')
    if len(url_parts) <= 1:
        styled_url = page.url
    else:
        for i in (-1, -2):
            if url_parts[i]:
                url_parts[i] = \
                    u'<span style="font-weight: bold">{0}</span>'.format(
                        url_parts[i]
                    )
                break
        styled_url = u'/'.join(url_parts)
    return mark_safe(
        u'<a href="{url}" style="font-weight: normal;">{text}</a>'.format(
            url=url, text=u'{url} &raquo;'.format(url=styled_url)
        )
    )


def sync_status(page):
    """Synchronization status of the Page"""
    if not page:
        return None
    status_parts = []
    if page.is_dirty:
        status_parts.append(
            u'<span style="color:black; font-weight:bold">'
            u'Changed in Admin!'
            u'</span>'
        )
    dump_status = PageFileDumper(page).status()
    if dump_status == SyncStatus.NO_CHANGES:
        dump_text = u'File is synced'
        dump_color = u'green'
    elif dump_status == SyncStatus.MODIFIED:
        dump_text = u'File content differs'
        dump_color = u'orange'
    elif dump_status == SyncStatus.ADDED:
        dump_text = u'File is missing'
        dump_color = u'red'
    else:  # Unknown sync status
        dump_text = u'?'
        dump_color = u'black'
    status_parts.append(
        u'<span style="color: {0}">{1}</span>'.format(
            dump_color, dump_text
        )
    )
    return mark_safe(u'<br>'.join(status_parts))


def save_page(page, user, created):
    """
    Marks the Page as "dirty" (modified in Admin),
    saves it and sends the "page edited" signal.
    """
    page.is_dirty = True
    page.save()
    page_edited.send(Page, page=page, user=user, created=created)


class PageAdmin(admin.ModelAdmin):
    """Admin interface options for Page model"""
    change_form_template = 'powerpages/admin/page_change_form.html'
    change_list_template = 'powerpages/admin/page_change_list.html'
    search_fields = (
        'url', 'alias', 'title', 'description', 'keywords', 'template'
    )
    list_display = (
        'url', 'alias', 'title', 'page_processor',
        'get_website_link', 'get_sync_status'
    )
    list_filter = ('page_processor',)
    readonly_fields = (
        'get_website_link', 'get_sync_status'
    )
    fieldsets = (
        (None, {
            'fields': (
                'get_website_link',
                'get_sync_status',
                'url',
                'alias',
            ),
        }),
        ('Metatags', {
            'fields': (
                'title',
                'description',
                'keywords',
            ),
        }),
        ('Template', {
            'fields': (
                'template',
            )
        }),
        ('Advanced', {
            'fields': (
                'page_processor',
                'page_processor_config',
            )
        }),
    )
    form = PageAdminForm

    def get_website_link(self, obj=None):
        return website_link(obj)
    get_website_link.short_description = u"URL"
    get_website_link.admin_order_field = 'url'
    get_website_link.allow_tags = True

    def get_sync_status(self, obj=None):
        return sync_status(obj)
    get_sync_status.short_description = u"Sync Status"
    get_sync_status.allow_tags = True

    def save_model(self, request, obj, form, change):
        save_page(page=obj, user=request.user, created=not change)


admin.site.register(Page, PageAdmin)



# -*- coding: utf-8 -*-

from django.core.cache import cache
from django.core.urlresolvers import reverse

from powerpages.cachekeys import url_cache
from powerpages.models import Page


def reverse_url(name, *args, **kwargs):
    """
    Looks for entry matching given name and optional parameters in cache
    and if it's not present tries to find CMS Page with given name,
    then tries to reverse given name with params to static URL.
    Updates the cache if result has been found and returns URL
    or throws exception.
    """
    # try to get url from cache
    url_key = url_cache(name, *args, **kwargs)
    url = cache.get(url_key, None)

    # try to find CMS Page only if optional params are omitted
    if not url and not args and not kwargs:
        try:
            page = Page.objects.get(alias=name)
            url = page.url
        except Page.DoesNotExist:
            pass

    # try to reverse url by the given name with arguments
    if not url:
        # reverse will throw exception if it fails
        url = reverse(name, args=args, kwargs=kwargs)

    cache.set(url_key, url)

    return url



# -*- coding: utf-8 -*-

import yaml

from django.test import TestCase

from powerpages.models import Page
from powerpages.forms import PageAdminForm


class PageFormTestCase(TestCase):

    maxDiff = None

    def test_valid_form_data_default_page_processor_root(self):
        data = {
            'url': '/',
            'alias': 'root',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.DefaultPageProcessor',
            'page_processor_config': '',
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.cleaned_data,
            {
                'url': '/',
                'alias': 'root',
                'description': 'At vero eos et accusamus et iusto odio',
                'keywords': 'lorem ipsum dolor sit amet',
                'page_processor': 'powerpages.DefaultPageProcessor',
                'page_processor_config': None,
                'template': '<h1>{{ website_page.title }}</h1>\n',
                'title': 'De Finibus Bonorum et Malorum'
            },
        )

    def test_valid_form_data_default_page_processor(self):
        Page.objects.create(url='/')
        data = {
            'url': '/test/',
            'alias': 'test-page',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.DefaultPageProcessor',
            'page_processor_config': '',
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertTrue(form.is_valid())
        self.assertDictEqual(
            form.cleaned_data,
            {
                'url': '/test/',
                'alias': 'test-page',
                'description': 'At vero eos et accusamus et iusto odio',
                'keywords': 'lorem ipsum dolor sit amet',
                'page_processor': 'powerpages.DefaultPageProcessor',
                'page_processor_config': None,
                'template': '<h1>{{ website_page.title }}</h1>\n',
                'title': 'De Finibus Bonorum et Malorum'
            }
        )

    def test_valid_form_data_redirect_processor(self):
        data = {
            'url': '/test/',
            'alias': 'test-page',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.RedirectProcessor',
            'page_processor_config': yaml.dump({
                'to url': '/test/'
            }),
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertTrue(form.is_valid())
        self.assertDictEqual(
            form.cleaned_data,
            {
                'url': '/test/',
                'alias': 'test-page',
                'description': 'At vero eos et accusamus et iusto odio',
                'keywords': 'lorem ipsum dolor sit amet',
                'page_processor': 'powerpages.RedirectProcessor',
                'page_processor_config': {
                    'to url': '/test/'
                },
                'template': '<h1>{{ website_page.title }}</h1>\n',
                'title': 'De Finibus Bonorum et Malorum'
            }
        )

    def test_valid_form_data_normalization(self):
        data = {
            'url': '/test/      ',
            'alias': 'test-page   ',
            'description': 'At vero eos et accusamus et iusto odio\n\n',
            'keywords': '   lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.RedirectProcessor',
            'page_processor_config': {
                'to url': '/test/'
            },
            'template': '\n\t<h1>{{ website_page.title }}</h1>\n   \n\n\r\n',
            'title': '  De Finibus Bonorum et Malorum  \t'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertTrue(form.is_valid())
        self.assertDictEqual(
            form.cleaned_data,
            {
                'url': '/test/',
                'alias': 'test-page',
                'description': 'At vero eos et accusamus et iusto odio',
                'keywords': 'lorem ipsum dolor sit amet',
                'page_processor': 'powerpages.RedirectProcessor',
                'page_processor_config': {
                    'to url': '/test/'
                },
                'template': '<h1>{{ website_page.title }}</h1>\n',
                'title': 'De Finibus Bonorum et Malorum'
            }
        )

    def test_invalid_form_data_unknown_url_name(self):
        data = {
            'url': '/test/',
            'alias': 'test-page',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.RedirectProcessor',
            'page_processor_config': {
                'to name': 'not-existing-url'
            },
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors.keys(), ['__all__'])

    def test_invalid_form_data_unknown_parent_template(self):
        # No parent page
        data = {
            'url': '/test/',
            'alias': 'test-page',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.DefaultPageProcessor',
            'page_processor_config': {
                'base template': "this-template-does-not-exist.html"
            },
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors.keys(), ['__all__'])

    def test_invalid_form_data_missing_url(self):
        # No parent page
        data = {
            'url': '',
            'alias': 'test-page',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.DefaultPageProcessor',
            'page_processor_config': None,
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors.keys(), ['url'])

    def test_invalid_form_data_duplicate_alias(self):
        Page.objects.create(url='/')
        Page.objects.create(url='/test-old/', alias='test-page')
        data = {
            'url': '/test/',
            'alias': 'test-page',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.DefaultPageProcessor',
            'page_processor_config': {},
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors.keys(), ['alias'])

    def test_invalid_form_data_malformed_yaml(self):
        data = {
            'url': '/test/',
            'alias': 'test-page',
            'description': 'At vero eos et accusamus et iusto odio',
            'keywords': 'lorem ipsum dolor sit amet',
            'page_processor': 'powerpages.DefaultPageProcessor',
            'page_processor_config': '- BAD [YAML CONFIG}',
            'template': '<h1>{{ website_page.title }}</h1>\n',
            'title': 'De Finibus Bonorum et Malorum'
        }
        form = PageAdminForm(data, instance=Page())
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors.keys(), ['page_processor_config'])



# -*- coding: utf-8 -*-

from django import forms
from django.test import RequestFactory
from django.contrib.auth.models import AnonymousUser

import yaml

from powerpages.models import Page
from powerpages.widgets import SourceCodeEditor
from powerpages.sync import normalize_page_fields


class PageAdminForm(forms.ModelForm):
    """
    ModelForm for Page's ModelAdmin
    """

    class Meta:
        model = Page
        exclude = ()
        widgets = {
            'template': SourceCodeEditor,
        }

    def clean_alias(self):
        """
        Force alias to be null instead of empty string
        (protection against breaking uniqueness constraint).
        """
        alias = self.cleaned_data.get('alias')
        if not alias:
            return None
        # uniqueness validation (not handled by default, because of null=True)
        page_query = Page.objects.all()
        if self.instance and self.instance.pk:
            page_query = page_query.exclude(pk=self.instance.pk)
        try:
            page = page_query.get(alias=alias)
        except Page.DoesNotExist:
            pass
        else:
            raise forms.ValidationError(
                'The "alias" field have to be unique or blank, but the same '
                'alias value has been found in page: %s.' % page
            )
        return alias

    def clean_page_processor_config(self):
        """Checks validity of Page processor Config, loads data from string"""
        value = (
            self.cleaned_data.get('page_processor_config') or ''
        ).strip() or None
        if value:
            try:
                value = yaml.load(value)
            except ValueError:
                raise forms.ValidationError('Invalid YAML config.')
            else:
                if not isinstance(value, dict):
                    raise forms.ValidationError('Invalid YAML config.')
        return value

    def clean(self):
        """Check validity of a Page template"""
        cleaned_data = self.cleaned_data
        if self._errors:
            return cleaned_data
        instance = self.instance
        normalized_data = cleaned_data.copy()
        normalized_data.update(normalize_page_fields(cleaned_data))
        # Copy original attributes for further restoring:
        original_attrs = {}
        for name, value in normalized_data.items():
            original_attrs[name] = getattr(instance, name)
            setattr(instance, name, value)
        # Do the validation
        request_factory = RequestFactory()
        request = request_factory.get(instance.url)
        request.session = {}
        request.user = AnonymousUser()
        try:
            processor = instance.get_page_processor()
            processor.validate(request=request)
        finally:
            # Restore instance to state existing before the validation:
            for name, value in original_attrs.items():
                setattr(instance, name, value)
        return normalized_data



# -*-coding: utf-8 -*-

import traceback

from django.forms import ValidationError
from django.core.urlresolvers import reverse
from django.core.cache import cache
from django.template import RequestContext, Context, Template
from django import http
from django.conf import settings
from django.utils.safestring import mark_safe

from powerpages.utils.class_registry.item import ConfigurableClassRegistryItem
from powerpages.utils.class_registry.config import ConfigVariable
from powerpages.settings import app_settings
from powerpages import page_processor_registry
from powerpages import cachekeys


def _get_cache_seconds(value):
    """`cache` config value converter"""
    return settings.CACHE_MIDDLEWARE_SECONDS if value is True else int(value)


class DefaultPageProcessor(ConfigurableClassRegistryItem):
    """Class responsible for rendering and validation of Pages."""

    CONFIG_VARIABLES = (
        ConfigVariable(
            'base template', converter=unicode,
            help_text="""
* name of template to be used in {% extends ... %},
* if not given parents' template is used and if page has no parent
- no base template is used.
            """,
        ),
        ConfigVariable(
            'cache', converter=_get_cache_seconds, default=0,
            help_text="""
* cache validity time in seconds,
* defaults to: `0` (no cache),
* if value is set to `true` default cache validity time is used."
(settings.CACHE_MIDDLEWARE_SECONDS).
            """,
        ),
        ConfigVariable(
            'cache for user', converter=bool, default=False,
            help_text="""
* default: `false`,
* if set to `true` page is cached per user (otherwise per language).
            """,
        ),
        ConfigVariable(
            'context processors', converter=list, default=list,
            help_text="""
* default: `[]`,
* list of context processor to be used in page template
(same as in settings.TEMPLATE_CONTEXT_PROCESSORS).
            """,
        ),
        ConfigVariable(
            'tag libraries', converter=list, default=list,
            help_text="""
* default: `[]`,
* list of tag libraries to be used in page template.
            """,
        ),
        ConfigVariable(
            'sitemap', default=True,
            help_text="""
* default: `true`
* if set to `true` page is visible in sitemap using default settings,
* if set to `false` page is not visible in sitemap,
* if set to dictionary, the following keys are used to modify
sitemap settings: `changefreq`, `lastmod`, `priority`.
            """,
        ),
        ConfigVariable(
            'headers', converter=dict, default=dict,
            help_text="""
* default: `{}`,
* dictionary with HTTP response headers.
            """,
        )
    )

    @property
    def page(self):
        """Explicit alias for model instance"""
        return self.model_instance

    def process_request(self, request, extra_context=None):
        """Main page processing logic"""
        context = self.get_rendering_context(request)
        if extra_context:
            context.update(extra_context)
        cache_key, seconds = self.get_cache_settings(request)
        if cache_key:
            content = cache.get(cache_key)
            if content is None:
                content = self.render(context)
                cache.set(cache_key, content, seconds)
        else:
            content = self.render(context)
        return self.create_response(request, content)

    def get_cache_settings(self, request):
        """
        Gives pair (cache_key, seconds) if cache is allowed for current page
        and double None otherwise
        """
        seconds = self.config.get('cache')
        if seconds:
            for_user = self.config.get('cache for user')
            if for_user:
                try:
                    user_id = request.user.id
                except AttributeError:
                    user_id = None
            else:
                user_id = None
            if user_id:
                cache_key = cachekeys.rendered_source_for_user(
                    self.page.pk, user_id
                )
            else:  # for_lang
                lang = request.LANGUAGE_CODE
                cache_key = cachekeys.rendered_source_for_lang(
                    self.page.pk, lang
                )
        else:
            cache_key, seconds = None, None
        return cache_key, seconds

    def render(self, context):
        """Render Page using given context"""
        source = self.get_template_source()
        page_template = Template(source)
        return page_template.render(context)

    def validate(self, request=None):
        """Check validity of configuration and Page template"""
        self.config.validate()
        try:
            context = self.get_validation_context(request=request)
            self.render(context)
        except:
            tb_info = traceback.format_exc()
            msg = u"""An error occurred trying to render the Page:
                <br/><pre>%s</pre>""" % tb_info
            raise ValidationError(mark_safe(msg))

    def create_response(self, request, content):
        """
        Creates HttpResponse and sets HTTP headers as defined in config.
        """
        response = http.HttpResponse(content)
        header_settings = self.config.get('headers')
        for key, value in header_settings.items():
            response[key] = value
        return response

    def get_extra_context(self, request):
        """
        Get custom context function from config and try to update
        RequestContext with it
        """
        extra_context = {}
        context_processors = self.config.get('context processors')
        for context_name in context_processors:
            context_name = context_name.split('.')
            context_module_name = ".".join(context_name[:-1])
            context_function_name = context_name[-1]
            try:
                context_module = __import__(
                    context_module_name, globals(), locals(),
                    [context_function_name, ], -1
                )
            except ImportError:
                pass
            else:
                context_function = getattr(
                    context_module, context_function_name
                )
                extra_context.update(context_function(request))
        return extra_context

    def create_request_context(self, request, data=None):
        """
        Create RequestContext for request and (optional) data.
        Update RequestContext with specified in config custom context function
        """
        data = data or {}
        extra_context = self.get_extra_context(request)
        data.update(extra_context)
        return RequestContext(request, data)

    def create_context(self, data=None):
        """Create RequestContext for request and (optional) data"""
        return Context(data)

    def get_validation_context(self, request=None):
        """Prepare context for template validation"""
        data = {}
        data['page'] = data['website_page'] = self.page
        data['page_processor'] = data['website_page_processor'] = self
        if request:
            context = self.create_request_context(request, data)
        else:
            context = self.create_context(data)
        return context

    def get_rendering_context(self, request):
        """Prepare context for page rendering"""
        data = {}
        data['page'] = data['website_page'] = self.page
        data['page_processor'] = data['website_page_processor'] = self
        return self.create_request_context(request, data)

    def get_extends_tag(self):
        """
        Prepares {% extends %} tag included
        at the beginning of template source.
        """
        base_template = self.config.get('base template')
        parent_page = self.page.parent()
        if parent_page and not base_template:
            base_template = "page/%s" % parent_page.pk
        if base_template:
            tag = u'{%% extends "%s" %%}' % base_template
        else:
            tag = u''
        return tag

    def get_load_tag(self):
        """
        Prepares {% load %} tag included at the beginning of template source.
        """
        tag_libraries = ['powerpages_tags']
        tag_libraries.extend(app_settings.TAG_LIBRARIES)
        tag_libraries.extend(self.config.get('tag libraries'))
        if tag_libraries:
            tag = u'{%% load %s %%}' % " ".join(tag_libraries)
        else:
            tag = u''
        return tag

    def get_template_source(self):
        """Prepare template source"""
        return (
            self.get_extends_tag() + self.get_load_tag() + self.page.template
        )

    def is_accessible(self):
        """Determines if Page can be accessed on URL using this processor"""
        return True


class RedirectProcessor(DefaultPageProcessor):
    """Class responsible for processing redirects."""

    CONFIG_VARIABLES = DefaultPageProcessor.CONFIG_VARIABLES + (
        ConfigVariable(
            'permanent', converter=bool, default=True,
            help_text="""
* default: `true`,
* defines if processor should use 301 or 302 HTTP response.
            """,
        ),
        ConfigVariable(
            'to alias', converter=unicode,
            help_text="""
* CMS alias of target Page.
            """,
        ),
        ConfigVariable(
            'to url', converter=unicode,
            help_text="""
* URL address of target page.
            """,
        ),
        ConfigVariable(
            'to name', converter=unicode,
            help_text="""
* name of target URL managed by Django url resolver,
* may be used together with `args` and `kwargs`.
            """,
        ),
        ConfigVariable(
            'args', converter=list, default=list,
            help_text="""
* default: `[]`,
* list of positional parameters to combine with `to name`.
            """,
        ),
        ConfigVariable(
            'kwargs', converter=dict, default=dict,
            help_text="""
* default: `{}`,
* dictionary of named parameters to combine with `to name`.
            """,
        ),
    )

    def process_request(self, request, extra_context=None):
        """Redirection processing logic"""
        try:
            location = self.get_redirect_location()
        except:
            location = '/'
            permanent = False
        else:
            permanent = self.config.get('permanent')
        response_class = (
            http.HttpResponsePermanentRedirect
            if permanent else
            http.HttpResponseRedirect
        )
        return response_class(location)

    def get_redirect_location(self):
        """Gets target location."""
        location_search_order = ('alias', 'url', 'name')
        location = None
        for method_postfix in location_search_order:
            provider = getattr(self, 'get_redirect_to_%s' % method_postfix)
            location = provider()
            if location:
                break
        return location or "/"

    def get_redirect_to_alias(self):
        """Location based on `alias` config option."""
        alias = self.config.get('to alias')
        if alias:
            page_model = self.page.__class__
            page = page_model.objects.get(alias=alias)
            return page.url

    def get_redirect_to_url(self):
        """Location based on `url` config option."""
        return self.config.get('to url')

    def get_redirect_to_name(self):
        """Location based on `name`, `args` and `kwargs` config options."""
        name = self.config.get('to name')
        if name:
            args = self.config.get('args')
            kwargs = self.config.get('kwargs')
            return reverse(name, args=args, kwargs=kwargs)

    def validate(self, request=None):
        try:
            self.get_redirect_location()
        except:
            tb_info = traceback.format_exc()
            msg = u"""An error occurred trying to render the Page:
            <br/><pre>%s</pre>""" % tb_info
            raise ValidationError(mark_safe(msg))


class NotFoundProcessor(DefaultPageProcessor):
    """
    Class responsible for 404 Page Not Found responses.
    Suitable for temporary deactivating selected pages.
    No special configuration needed.
    """

    def process_request(self, request, extra_context=None):
        """
        Raises Http404 to use django's builtin HTTP 404 Response.
        """
        raise http.Http404()

    def validate(self, request=None):
        """No need to validate anything here"""
        pass

    def is_accessible(self):
        """Determines if Page can be accessed on URL using this processor"""
        return False


page_processor_registry.register(DefaultPageProcessor)
page_processor_registry.register(RedirectProcessor)
page_processor_registry.register(NotFoundProcessor)



# -*- coding: utf-8 -*-


from powerpages.utils.class_registry.config import Config


class ConfigurableClassRegistryItem(object):
    """Base class for classes stored in registry which may be used"""

    CONFIG_VARIABLES = ()

    def __init__(self, model_instance, config):
        self.model_instance = model_instance
        self.config = Config(self.CONFIG_VARIABLES, config)



# -*- coding: utf-8 -*-

from django import template
from django.template.base import kwarg_re
from django.utils.encoding import smart_str

from powerpages.reverse import reverse_url


register = template.Library()


def page_url_compiler(node_class):
    """
    Creates compile function for page_url* tag family
    {% tag_name alias_or_name value1 value2 kwarg1=value3 kwarg2=value4 %}
    """

    def _page_url_parser(parser, token):
        """
        Inserts page URL for given alias.
        If Page instance with given alias is not found in the CMS,
        tag acts like a standard Django {% url %} tag.
        """

        bits = token.split_contents()
        if len(bits) < 2:
            raise template.TemplateSyntaxError(
                "'%s' takes at least one argument (path to a view)" % bits[0]
            )
        name = bits[1]
        args = []
        kwargs = {}
        as_var = None
        bits = bits[2:]

        # handling "as" functionality
        if len(bits) >= 2 and bits[-2] == 'as':
            as_var = bits[-1]
            bits = bits[:-2]

        # processing bits as template vars
        if len(bits):
            for bit in bits:
                match = kwarg_re.match(bit)
                if not match:
                    raise template.TemplateSyntaxError(
                        "Malformed arguments to url tag"
                    )
                arg_name, arg_value = match.groups()
                if arg_name:
                    kwargs[arg_name] = parser.compile_filter(arg_value)
                else:
                    args.append(parser.compile_filter(arg_value))

        return node_class(name, args, kwargs, as_var)

    return _page_url_parser


class PageURLNode(template.Node):
    """
    Template tag for inserting URL leading to CMS Page or static URL.
    """

    def __init__(self, name, args, kwargs, as_var):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.as_var = as_var

    def render(self, context):
        args = [arg.resolve(context) for arg in self.args]
        kwargs = dict([(smart_str(k, 'ascii'), v.resolve(context))
                       for k, v in self.kwargs.items()])

        name = self.name
        url = self.get_url(name, *args, **kwargs)

        if self.as_var:
            context[self.as_var] = url
            return ''
        else:
            return url

    def get_url(self, name, *args, **kwargs):
        """URL generation"""
        return reverse_url(name, *args, **kwargs)


page_url = register.tag(
    'page_url', page_url_compiler(PageURLNode)
)



# -*- coding: utf-8 -*-


def cache_result_on(attribute, as_property=False):
    """
    Decorator for methods whose value is cached on given attribute.
    `attribute` is a string template and method parameters can be used
    to render it.
    `as_property` can be used to convert method into property.
    """
    def _cache_result_on(method):
        def __cache_result_on(self, *args, **kwargs):
            cached_attr = attribute.format(*args, **kwargs)
            if not hasattr(self, cached_attr):
                setattr(self, cached_attr, method(self, *args, **kwargs))
            return getattr(self, cached_attr)
        if as_property:
            __cache_result_on = property(__cache_result_on)
        return __cache_result_on
    return _cache_result_on



# -*- coding: utf-8 -*-

from django.core.urlresolvers import reverse

from powerpages.settings import app_settings


class NamedURL(object):
    """Wrapper over named URLs to provide lazy reversion"""

    def __init__(self, name, *args, **kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs

    def reverse_url(self):
        """Reverses given URL"""
        return reverse(self.name, args=self.args, kwargs=self.kwargs)


class Sitemap(object):
    """Single sitemap configuration class."""

    protocol = None
    domain = None
    changefreq = None
    priority = None

    class URL(object):
        """Single sitemap URL"""

        def __init__(self, location, lastmod=None, changefreq=None,
                     priority=None):
            self.location = location
            self.lastmod = lastmod
            self.changefreq = changefreq
            self.priority = priority

    def __init__(self, request=None):
        if self.protocol is None:
            if app_settings.SITEMAP_PROTOCOL:
                self.protocol = app_settings.SITEMAP_PROTOCOL
            elif request:
                self.protocol = 'https' if request.is_secure() else 'http'
            else:
                self.protocol = 'http'
        if self.domain is None:
            if app_settings.SITEMAP_DOMAIN:
                self.domain = app_settings.SITEMAP_DOMAIN
            elif request:
                self.domain = request.get_host()
            else:
                self.domain = 'localhost'
        if self.changefreq is None:
            self.changefreq = app_settings.SITEMAP_DEFAULT_CHANGEFREQ
        if self.priority is None:
            self.priority = app_settings.SITEMAP_DEFAULT_PRIORITY

    def get_items(self):
        """To be overriden when dynamic items needed"""
        return getattr(self, 'items', ())

    def get_urls(self):
        """Generator over all URL instances provided by this sitemap."""
        for conf_item in self.get_items():
            url_kwargs = conf_item.copy()
            raw_location = url_kwargs.pop('location')
            if hasattr(raw_location, 'reverse_url'):
                raw_location = raw_location.reverse_url()
            location = "%s://%s%s" % (
                self.protocol, self.domain, raw_location
            )
            if 'changefreq' not in url_kwargs and self.changefreq:
                url_kwargs['changefreq'] = self.changefreq
            if 'priority' not in url_kwargs and self.priority:
                url_kwargs['priority'] = self.priority
            url = self.URL(location, **url_kwargs)
            yield url


class ModelSitemap(Sitemap):
    """
    Single sitemap configuration class with items based on model
    instances from given queryset.
    Subclasses have to contain `queryset` attribute.
    """

    def get_items(self):
        for obj in self.queryset.all():  # working on queryset copy
            yield self.from_instance(obj)

    def get_instance_location(self, obj):
        """Retrieves URL of particular model instance."""
        return obj.get_absolute_url()

    def from_instance(self, obj):
        """Converts model instance into dictionary."""
        return {
            'location': self.get_instance_location(obj),
        }


class URLSet(object):
    """Container for Sitemaps"""

    def __init__(self):
        self.sitemaps = set()

    def add(self, sitemap):
        self.sitemaps.add(sitemap)

    def urls(self, request=None):
        """Generator over all URL instances provided by all sitemaps."""
        for sitemap_class in self.sitemaps:
            sitemap = sitemap_class(request=request)
            for url in sitemap.get_urls():
                yield url


sitemaps = URLSet()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import tempfile
import shutil
import codecs
import StringIO

from django.test import TestCase
from django.test.utils import override_settings

from powerpages.models import Page
from powerpages.sync import (
    PageFileDumper, FilePageLoader, SyncStatus,
    WebsiteDumpOperation, WebsiteLoadOperation, normalize_page_fields
)


def _file_contents(s, strip_spaces=True):
    s = s.replace(',\n', ', \n')
    return s.strip(' ') if strip_spaces else s


class BaseSyncTestCase(TestCase):
    normalized_content = _file_contents('''{
  "alias": "test-page",
  "description": "At vero eos et accusamus et iusto odio",
  "keywords": "lorem ipsum dolor sit amet",
  "page_processor": "powerpages.RedirectProcessor",
  "page_processor_config": {
    "to url": "/test/"
  },
  "title": "De Finibus Bonorum et Malorum"
}
## TEMPLATE SOURCE: ##
<h1>{{ website_page.title }}</h1>
    ''')

    edited_content = '''
{
      "alias":    "test-page",
      "description": "At vero eos et accusamus et iusto odio",
      "title": "De Finibus Bonorum et Malorum",

      "page_processor_config": {"to url": "/test/"},
      "page_processor": "powerpages.RedirectProcessor",
      "keywords": "lorem ipsum dolor sit amet"
}
## TEMPLATE SOURCE: ##
<h1>{{ website_page.title }}</h1>'''

    simple_content = _file_contents('''{
  "description": "",
  "title": "",
  "alias": null,
  "page_processor_config": null,
  "page_processor": "powerpages.DefaultPageProcessor",
  "keywords": ""
}
## TEMPLATE SOURCE: ##
    ''')

    def setUp(self):
        self.sync_directory = tempfile.mkdtemp()
        self.settings_change = override_settings(
            POWER_PAGES={'SYNC_DIRECTORY': self.sync_directory}
        )
        self.settings_change.enable()

    def tearDown(self):
        self.settings_change.disable()
        shutil.rmtree(self.sync_directory)

    def _make_file(self, relative_path, content, make_dirs=True):
        absolute_path = os.path.join(self.sync_directory, relative_path)
        if make_dirs:
            absolute_dir_path, file_name = os.path.split(absolute_path)
            os.makedirs(absolute_dir_path)
        with codecs.open(absolute_path, 'w', encoding='utf-8') as f:
            f.write(content)


class PageFileDumperTestCase(BaseSyncTestCase):

    def test_relative_path_root(self):
        page = Page.objects.create(
            url='/',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(dumper.relative_path(), '_index_.page')

    def test_relative_path_nested(self):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(dumper.relative_path(), 'a/b/test.page')

    def test_relative_path_txt(self):
        page = Page.objects.create(
            url='robots.txt',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(dumper.relative_path(), 'robots.txt')

    def test_absolute_path(self):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(
            dumper.absolute_path(),
            os.path.join(self.sync_directory, 'a/b/test.page')
        )

    def test_page_fields_default(self):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(
            dumper.page_fields(),
            {
                'alias': None,
                'description': '',
                'keywords': '',
                'page_processor': 'powerpages.DefaultPageProcessor',
                'page_processor_config': None,
                'template': '',
                'title': ''
            }
        )

    def test_page_fields_full(self):
        page = Page.objects.create(
            url='/a/b/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(
            dumper.page_fields(),
            {
                'alias': 'test-page',
                'description': 'At vero eos et accusamus et iusto odio',
                'keywords': 'lorem ipsum dolor sit amet',
                'page_processor': 'powerpages.RedirectProcessor',
                'page_processor_config': {
                    'to url': '/test/'
                },
                'template': '<h1>{{ website_page.title }}</h1>\n',
                'title': 'De Finibus Bonorum et Malorum'
            }
        )

    def test_file_contents_default(self):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(
            dumper.file_contents(),
            _file_contents('''{
  "alias": null,
  "description": "",
  "keywords": "",
  "page_processor": "powerpages.DefaultPageProcessor",
  "page_processor_config": null,
  "title": ""
}
## TEMPLATE SOURCE: ##
            ''')
        )

    def test_file_contents_full(self):
        page = Page.objects.create(
            url='/a/b/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(
            dumper.file_contents(),
            _file_contents('''{
  "alias": "test-page",
  "description": "At vero eos et accusamus et iusto odio",
  "keywords": "lorem ipsum dolor sit amet",
  "page_processor": "powerpages.RedirectProcessor",
  "page_processor_config": {
    "to url": "/test/"
  },
  "title": "De Finibus Bonorum et Malorum"
}
## TEMPLATE SOURCE: ##
<h1>{{ website_page.title }}</h1>
            ''')
        )

    def test_file_exists_false(self):
        page = Page.objects.create(
            url='/',
        )
        dumper = PageFileDumper(page)
        self.assertFalse(dumper.file_exists())

    def test_file_exists_true(self):
        page = Page.objects.create(
            url='/',
        )
        open(os.path.join(self.sync_directory, '_index_.page'), 'a').close()
        dumper = PageFileDumper(page)
        self.assertTrue(dumper.file_exists())

    def test_status_added(self):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        dumper = PageFileDumper(page)
        self.assertEqual(dumper.status(), SyncStatus.ADDED)

    def test_status_no_changes(self):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        PageFileDumper(page).save()
        dumper = PageFileDumper(page)
        self.assertEqual(dumper.status(), SyncStatus.NO_CHANGES)

    def test_status_no_changes_saved(self):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        PageFileDumper(page).save()
        page.save()
        dumper = PageFileDumper(page)
        self.assertEqual(dumper.status(), SyncStatus.NO_CHANGES)

    def _test_status_modified(self, **kwargs):
        page = Page.objects.create(
            url='/a/b/test/',
        )
        PageFileDumper(page).save()
        for attr, value in kwargs.items():
            setattr(page, attr, value)
        page.save()
        dumper = PageFileDumper(page)
        self.assertEqual(dumper.status(), SyncStatus.MODIFIED)

    def test_status_modified_template(self):
        self._test_status_modified(template='<h1>TEST</h1>\n')

    def test_status_modified_title(self):
        self._test_status_modified(title='TEST')

    def test_status_modified_description(self):
        self._test_status_modified(description='Test test test.')

    def test_status_modified_keywords(self):
        self._test_status_modified(keywords='test test test')

    def test_status_modified_alias(self):
        self._test_status_modified(alias='test-page')

    def test_status_modified_page_processor(self):
        self._test_status_modified(
            page_processor='powerpages.RedirectProcessor'
        )

    def test_status_modified_page_processor_config(self):
        self._test_status_modified(page_processor_config={'sitemap': False})

    def test_save_file_contents(self):
        page = Page.objects.create(
            url='/a/b/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        PageFileDumper(page).save()
        path = os.path.join(self.sync_directory, 'a/b/test.page')
        self.assertTrue(os.path.exists(path))
        with open(path) as f:
            file_contents = f.read()
        self.assertEqual(
            file_contents,
            _file_contents('''{
  "alias": "test-page",
  "description": "At vero eos et accusamus et iusto odio",
  "keywords": "lorem ipsum dolor sit amet",
  "page_processor": "powerpages.RedirectProcessor",
  "page_processor_config": {
    "to url": "/test/"
  },
  "title": "De Finibus Bonorum et Malorum"
}
## TEMPLATE SOURCE: ##
<h1>{{ website_page.title }}</h1>
            ''')
        )

    def test_save_is_dirty(self):
        page = Page.objects.create(
            url='/a/b/test/',
            is_dirty=True
        )
        PageFileDumper(page).save()
        page = Page.objects.get(pk=page.pk)
        self.assertFalse(page.is_dirty)

    def test_diff_template_modified(self):
        page = Page.objects.create(
            url='/a/b/test/',
            template="<p>\nTEST\n<\p>\n"
        )
        PageFileDumper(page).save()
        page.template = "<p>\nCHANGE\n<\p>\n"
        page.save()
        dumper = PageFileDumper(page)
        self.assertEqual(
            dumper.diff(),
            '''--- Current content

+++ Coming changes

@@ -8,5 +8,5 @@

 }
 ## TEMPLATE SOURCE: ##
 <p>
-TEST
+CHANGE
 <\p>'''
        )

    def test_diff_title_modified(self):
        page = Page.objects.create(
            url='/a/b/test/',
            title="TEST"
        )
        PageFileDumper(page).save()
        page.title = "CHANGE"
        page.save()
        dumper = PageFileDumper(page)
        self.assertEqual(
            dumper.diff(),
            _file_contents('''--- Current content

+++ Coming changes

@@ -4,6 +4,6 @@

   "keywords": "",
   "page_processor": "powerpages.DefaultPageProcessor",
   "page_processor_config": null,
-  "title": "TEST"
+  "title": "CHANGE"
 }
 ## TEMPLATE SOURCE: ##''')
        )


class PageFileLoaderTestCase(BaseSyncTestCase):

    def test_data_normalization(self):
        self.assertEqual(
            normalize_page_fields(
                FilePageLoader.load(self.normalized_content)
            ),
            normalize_page_fields(
                FilePageLoader.load(self.edited_content)
            )
        )

    def test_url_root(self):
        loader = FilePageLoader('_index_.page')
        self.assertEqual(loader.url(), '/')

    def test_url_nested(self):
        loader = FilePageLoader('a/b/test.page')
        self.assertEqual(loader.url(), '/a/b/test/')

    def test_relative_path_txt(self):
        loader = FilePageLoader('robots.txt')
        self.assertEqual(loader.url(), '/robots.txt')

    def test_absolute_path(self):
        loader = FilePageLoader('a/b/test.page')
        self.assertEqual(
            loader.absolute_path(),
            os.path.join(self.sync_directory, 'a/b/test.page')
        )

    def test_page_fields_normalized(self):
        path = 'a/b/test.page'
        self._make_file(path, self.normalized_content)
        loader = FilePageLoader(path)
        self.assertEqual(
            loader.page_fields(),
            {
                'alias': 'test-page',
                'description': 'At vero eos et accusamus et iusto odio',
                'keywords': 'lorem ipsum dolor sit amet',
                'page_processor': 'powerpages.RedirectProcessor',
                'page_processor_config': {
                    'to url': '/test/'
                },
                'template': '<h1>{{ website_page.title }}</h1>\n',
                'title': 'De Finibus Bonorum et Malorum'
            }
        )

    def test_page_fields_edited(self):
        path = 'a/b/test.page'
        self._make_file(path, self.edited_content)
        loader = FilePageLoader(path)
        self.assertEqual(
            loader.page_fields(),
            {
                'alias': 'test-page',
                'description': 'At vero eos et accusamus et iusto odio',
                'keywords': 'lorem ipsum dolor sit amet',
                'page_processor': 'powerpages.RedirectProcessor',
                'page_processor_config': {
                    'to url': '/test/'
                },
                'template': '<h1>{{ website_page.title }}</h1>\n',
                'title': 'De Finibus Bonorum et Malorum'
            }
        )

    def test_file_contents(self):
        path = 'a/b/test.page'
        self._make_file(path, self.edited_content)
        loader = FilePageLoader(path)
        self.assertEqual(loader.file_contents(), self.edited_content)

    def test_page_existing(self):
        page = Page.objects.create(url='/a/b/test/')
        path = 'a/b/test.page'
        self._make_file(path, self.normalized_content)
        loader = FilePageLoader(path)
        self.assertEqual(loader.page(), page)

    def test_page_non_existing(self):
        path = 'a/b/test.page'
        self._make_file(path, self.normalized_content)
        loader = FilePageLoader(path)
        self.assertIsNone(loader.page())

    def test_status_added(self):
        path = 'a/b/test.page'
        self._make_file(path, self.edited_content)
        loader = FilePageLoader(path)
        self.assertEqual(loader.status(), SyncStatus.ADDED)

    def test_status_no_changes(self):
        path = 'a/b/test.page'
        self._make_file(path, self.edited_content)
        FilePageLoader(path).save()
        loader = FilePageLoader(path)
        self.assertEqual(loader.status(), SyncStatus.NO_CHANGES)

    def _test_status_modified(self, **kwargs):
        path = 'a/b/test.page'
        self._make_file(path, self.edited_content)
        FilePageLoader(path).save()
        page = Page.objects.get(url='/a/b/test/')
        for attr, value in kwargs.items():
            setattr(page, attr, value)
        page.save()
        loader = FilePageLoader(path)
        self.assertEqual(loader.status(), SyncStatus.MODIFIED)

    def test_status_modified_template(self):
        self._test_status_modified(template='<h1>TEST</h1>')

    def test_status_modified_title(self):
        self._test_status_modified(title='TEST')

    def test_status_modified_description(self):
        self._test_status_modified(description='Test test test.')

    def test_status_modified_keywords(self):
        self._test_status_modified(keywords='test test test')

    def test_status_modified_alias(self):
        self._test_status_modified(alias='test-page-2')

    def test_status_modified_page_processor(self):
        self._test_status_modified(
            page_processor='powerpages.NotFoundProcessor'
        )

    def test_status_modified_page_processor_config(self):
        self._test_status_modified(page_processor_config={'sitemap': False})

    def test_diff_template_modified(self):
        path = 'a/b/test.page'
        self._make_file(path, self.edited_content)
        FilePageLoader(path).save()
        page = Page.objects.get(url='/a/b/test/')
        page.template = 'TEST'
        page.save()
        loader = FilePageLoader(path)
        self.assertEqual(
            loader.diff(),
            _file_contents('''--- Current content

+++ Coming changes

@@ -9,4 +9,4 @@

   "title": "De Finibus Bonorum et Malorum"
 }
 ## TEMPLATE SOURCE: ##
-TEST
+<h1>{{ website_page.title }}</h1>''')
        )

    def test_diff_title_modified(self):
        path = 'a/b/test.page'
        self._make_file(path, self.edited_content)
        FilePageLoader(path).save()
        page = Page.objects.get(url='/a/b/test/')
        page.title = 'CHANGE'
        page.save()
        loader = FilePageLoader(path)
        self.assertEqual(
            loader.diff(),
            _file_contents('''--- Current content

+++ Coming changes

@@ -6,7 +6,7 @@

   "page_processor_config": {
     "to url": "/test/"
   },
-  "title": "CHANGE"
+  "title": "De Finibus Bonorum et Malorum"
 }
 ## TEMPLATE SOURCE: ##
 <h1>{{ website_page.title }}</h1>''')
        )


class WebsiteDumpOperationTestCase(BaseSyncTestCase):

    def test_dump_new_file(self):
        Page.objects.create(
            url='/a/b/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        stdout = StringIO.StringIO()
        stderr = StringIO.StringIO()
        operation = WebsiteDumpOperation(
            root_url='/a/b/test/',
            stdout=stdout,
            stderr=stderr,
            get_input=lambda p: 'y',
            dry_run=False,
            no_interactive=False,
            quiet=False,
            force=False,
            git_add=False,
            no_color=True,
        )
        operation.run()
        # Check result file:
        path = os.path.join(self.sync_directory, 'a/b/test.page')
        self.assertTrue(os.path.exists(path))
        with codecs.open(path, encoding='utf-8') as f:
            file_contents = f.read()
        self.assertEqual(
            file_contents,
            _file_contents('''{
  "alias": "test-page",
  "description": "At vero eos et accusamus et iusto odio",
  "keywords": "lorem ipsum dolor sit amet",
  "page_processor": "powerpages.RedirectProcessor",
  "page_processor_config": {
    "to url": "/test/"
  },
  "title": "De Finibus Bonorum et Malorum"
}
## TEMPLATE SOURCE: ##
<h1>{{ website_page.title }}</h1>
            ''')
        )
        output = stdout.getvalue()
        # Check stdout:
        self.assertIn('[A] = 1', output)  # 1 file created
        self.assertNotIn('[M] = ', output)  # no modifications
        self.assertNotIn('[D] = ', output)  # no deletions
        # Check stderr:
        self.assertEqual(stderr.getvalue(), '')  # no errors

    def test_dump_modified_file(self):
        page = Page.objects.create(
            url='/a/b/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        PageFileDumper(page).save()  # file is created
        page.title = "CHANGE!"
        page.save()
        stdout = StringIO.StringIO()
        stderr = StringIO.StringIO()
        operation = WebsiteDumpOperation(
            root_url='/a/b/test/',
            stdout=stdout,
            stderr=stderr,
            get_input=lambda p: 'y',
            dry_run=False,
            no_interactive=False,
            quiet=False,
            force=False,
            git_add=False,
            no_color=True,
        )
        operation.run()
        # Check result file:
        path = os.path.join(self.sync_directory, 'a/b/test.page')
        self.assertTrue(os.path.exists(path))
        with codecs.open(path, encoding='utf-8') as f:
            file_contents = f.read()
        self.assertEqual(
            file_contents,
            _file_contents('''{
  "alias": "test-page",
  "description": "At vero eos et accusamus et iusto odio",
  "keywords": "lorem ipsum dolor sit amet",
  "page_processor": "powerpages.RedirectProcessor",
  "page_processor_config": {
    "to url": "/test/"
  },
  "title": "CHANGE!"
}
## TEMPLATE SOURCE: ##
<h1>{{ website_page.title }}</h1>
            ''')
        )
        output = stdout.getvalue()
        # Check stdout:
        self.assertNotIn('[A] = ', output)  # no additions
        self.assertIn('[M] = 1', output)  # 1 file modified
        self.assertNotIn('[D] = ', output)  # no deletions
        # Check stderr:
        self.assertEqual(stderr.getvalue(), '')  # no errors

    def test_dump_deleted_file(self):
        # root page:
        root = Page.objects.create(url='/')
        PageFileDumper(root).save()
        # sibling page:
        sibling = Page.objects.create(url='/dummy/')
        PageFileDumper(sibling).save()
        # TODO: handle removing all children

        page = Page.objects.create(
            url='/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        PageFileDumper(page).save()  # file is created
        page.delete()
        stdout = StringIO.StringIO()
        stderr = StringIO.StringIO()
        operation = WebsiteDumpOperation(
            root_url='/',
            stdout=stdout,
            stderr=stderr,
            get_input=lambda p: 'y',
            dry_run=False,
            no_interactive=False,
            quiet=False,
            force=False,
            git_add=False,
            no_color=True,
        )
        operation.run()
        # Check result file:
        path = os.path.join(self.sync_directory, 'test.page')
        self.assertFalse(os.path.exists(path))
        output = stdout.getvalue()
        # Check stdout:
        self.assertNotIn('[A] = ', output)  # no additions
        self.assertNotIn('[M] = ', output)  # no modifications
        self.assertIn('[D] = 1', output)  # 1 file deleted
        # Check stderr:
        self.assertEqual(stderr.getvalue(), '')  # no errors


class WebsiteLoadOperationTestCase(BaseSyncTestCase):

    def test_load_new_page(self):
        # root page:
        root = Page.objects.create(url='/')
        PageFileDumper(root).save()
        # sibling page:
        sibling = Page.objects.create(url='/dummy/')
        PageFileDumper(sibling).save()

        self._make_file(
            'test.page',
            _file_contents('''{
  "alias": "test-page",
  "description": "At vero eos et accusamus et iusto odio",
  "keywords": "lorem ipsum dolor sit amet",
  "page_processor": "powerpages.RedirectProcessor",
  "page_processor_config": {
    "to url": "/test/"
  },
  "title": "De Finibus Bonorum et Malorum"
}
## TEMPLATE SOURCE: ##
<h1>{{ website_page.title }}</h1>
            '''),
            make_dirs=False
        )

        stdout = StringIO.StringIO()
        stderr = StringIO.StringIO()
        operation = WebsiteLoadOperation(
            root_url='/',
            stdout=stdout,
            stderr=stderr,
            get_input=lambda p: 'y',
            dry_run=False,
            no_interactive=False,
            quiet=False,
            force=False,
            git_add=False,
            no_color=True,
        )
        operation.run()
        # Check result file:
        page = Page.objects.filter(url='/test/').first()
        self.assertIsNotNone(page)
        self.assertEqual(page.alias, "test-page")
        self.assertEqual(
            page.description, "At vero eos et accusamus et iusto odio"
        )
        self.assertEqual(page.keywords, "lorem ipsum dolor sit amet")
        self.assertEqual(page.page_processor, "powerpages.RedirectProcessor")
        self.assertEqual(page.page_processor_config, {"to url": "/test/"})
        self.assertEqual(page.title, "De Finibus Bonorum et Malorum")
        self.assertEqual(page.template, '<h1>{{ website_page.title }}</h1>\n')

        output = stdout.getvalue()
        # Check stdout:
        self.assertIn('[A] = 1', output)  # 1 page added
        self.assertNotIn('[M] = ', output)  # no modifications
        self.assertNotIn('[D] = ', output)  # no deletions
        # Check stderr:
        self.assertEqual(stderr.getvalue(), '')  # no errors

    def test_load_modified_page(self):
        # root page:
        root = Page.objects.create(url='/')
        PageFileDumper(root).save()
        # sibling page:
        sibling = Page.objects.create(url='/dummy/')
        PageFileDumper(sibling).save()

        Page.objects.create(
            url='/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        self._make_file(
            'test.page',
            _file_contents('''{
  "alias": "test-page",
  "description": "At vero eos et accusamus et iusto odio",
  "keywords": "lorem ipsum dolor sit amet",
  "page_processor": "powerpages.RedirectProcessor",
  "page_processor_config": {
    "to url": "/test/"
  },
  "title": "TEST"
}
## TEMPLATE SOURCE: ##
<h1>CHANGE</h1>
            '''),
            make_dirs=False
        )

        stdout = StringIO.StringIO()
        stderr = StringIO.StringIO()
        operation = WebsiteLoadOperation(
            root_url='/',
            stdout=stdout,
            stderr=stderr,
            get_input=lambda p: 'y',
            dry_run=False,
            no_interactive=False,
            quiet=False,
            force=False,
            git_add=False,
            no_color=True,
        )
        operation.run()
        # Check result file:
        page = Page.objects.filter(url='/test/').first()
        self.assertIsNotNone(page)
        self.assertEqual(page.alias, "test-page")
        self.assertEqual(
            page.description, "At vero eos et accusamus et iusto odio"
        )
        self.assertEqual(page.keywords, "lorem ipsum dolor sit amet")
        self.assertEqual(page.page_processor, "powerpages.RedirectProcessor")
        self.assertEqual(page.page_processor_config, {"to url": "/test/"})
        self.assertEqual(page.title, "TEST")
        self.assertEqual(page.template, '<h1>CHANGE</h1>\n')

        output = stdout.getvalue()
        # Check stdout:
        self.assertNotIn('[A] = ', output)  # no additions
        self.assertIn('[M] = 1', output)  # 1 page modified
        self.assertNotIn('[D] = ', output)  # no deletions
        # Check stderr:
        self.assertEqual(stderr.getvalue(), '')  # no errors

    def test_load_deleted_page(self):
        # root page:
        root = Page.objects.create(url='/')
        PageFileDumper(root).save()
        # sibling page:
        sibling = Page.objects.create(url='/dummy/')
        PageFileDumper(sibling).save()

        Page.objects.create(
            url='/test/',
            alias='test-page',
            description='At vero eos et accusamus et iusto odio',
            keywords='lorem ipsum dolor sit amet',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={
                'to url': '/test/'
            },
            template='<h1>{{ website_page.title }}</h1>\n',
            title='De Finibus Bonorum et Malorum',
        )
        # no file!

        stdout = StringIO.StringIO()
        stderr = StringIO.StringIO()
        operation = WebsiteLoadOperation(
            root_url='/',
            stdout=stdout,
            stderr=stderr,
            get_input=lambda p: 'y',
            dry_run=False,
            no_interactive=False,
            quiet=False,
            force=False,
            git_add=False,
            no_color=True,
        )
        operation.run()
        # Check result file:
        page = Page.objects.filter(url='/test/').first()
        self.assertIsNone(page)

        output = stdout.getvalue()
        # Check stdout:
        self.assertNotIn('[A] = ', output)  # no additions
        self.assertNotIn('[M] = ', output)  # no modifications
        self.assertIn('[D] = 1', output)  # 1 page deleted
        # Check stderr:
        self.assertEqual(stderr.getvalue(), '')  # no errors



# -*- coding: utf-8 -*-

from django.conf.urls import include, url
from django.contrib import admin


urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(r'', include('powerpages.urls')),
]



# -*- coding: utf-8 -*-

from .website_base import BaseDumpLoadCommand
from powerpages.sync import WebsiteDumpOperation


class Command(BaseDumpLoadCommand):
    """DUMPS website Pages TO structure of directories and file"""

    operation_class = WebsiteDumpOperation



# -*- coding: utf-8 -*-

"""Bunch of tools to provide nice console output."""

import sys
import math
import re
import datetime
import itertools
import threading


_thread_locals = threading.local()


def total_seconds(time_delta):
    """
    Total number of seconds if given datetime.timedelta object.
    Required due to missing datetime.timedelta.total_seconds() in py26.
    """
    return (
        time_delta.days * 24 * 60 * 60 +
        time_delta.seconds +
        time_delta.microseconds / 1000000.0
    )


class BaseStyle(object):
    """Allows to apply colors to texts printed on console"""

    RESET = ''

    @classmethod
    def style(cls, option, text):
        """Applies styling option to given text"""
        return u"{start}{text}{reset}".format(
            start=getattr(cls, option), text=text, reset=cls.RESET
        )


class ForegroundColor(BaseStyle):
    """Container for foreground console colors (ANSI)"""
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    RESET = '\033[39m'


class BackgroundColor(BaseStyle):
    """Container for background console colors (ANSI)"""
    BLACK = '\033[40m'
    RED = '\033[41m'
    GREEN = '\033[42m'
    YELLOW = '\033[43m'
    BLUE = '\033[44m'
    MAGENTA = '\033[45m'
    CYAN = '\033[46m'
    WHITE = '\033[47m'
    RESET = '\033[49m'


class Font(BaseStyle):
    """Container for font styling (ANSI)"""
    BOLD = '\033[2m'
    RESET = '\033[0m'


class Console(object):
    """Console writing utilities"""

    def __init__(self, stream=None):
        if stream is None:
            stream = sys.stdout
        self.stream = stream

    def format(self, text, template=u"{text:<{min_length}}",
               bg_color=None, fg_color=None, font=None, min_length=80):
        """Formats the output string"""
        if bg_color:
            text = BackgroundColor.style(bg_color, text)
        if fg_color:
            text = ForegroundColor.style(fg_color, text)
        if font:
            text = Font.style(font, text)
        return template.format(text=text, min_length=min_length)

    def write(self, text):
        """Writes text to the output stream"""
        self.stream.write(text)
        self.stream.flush()

    def same_line(self, text, **format_options):
        """Writes text in the same line"""
        format_options['template'] = u'\r{text:<{min_length}}'
        self.write(self.format(text, **format_options))

    def new_line(self, text,  **format_options):
        """Writes text in the new line"""
        format_options['template'] = u'\r{text:<{min_length}}\n'
        self.write(self.format(text, **format_options))

    # Shortcuts:

    def info(self, message):
        """Writes non-formatted line of text into output stream"""
        self.new_line(message)

    def warning(self, message):
        """Writes yellow line of text into output stream"""
        self.new_line(message, fg_color='YELLOW')

    def failure(self, message):
        """Writes red line of text into output stream"""
        self.new_line(message, fg_color='RED')

    def success(self, message):
        """Writes green line of text into output stream"""
        self.new_line(message, fg_color='GREEN')


class ProgressBar(object):
    """Console progress bar"""
    TEMPLATE = '$progress $percent ($current / $total) $eta $screw'
    PROGRESS_TEMPLATE = '[{bricks: <{num_bricks}}]'
    PROGRESS_BRICK = '#'
    PROGRESS_NUM_BRICKS = 20
    PERCENT_TEMPLATE = '{ratio:7.2%}'
    CURRENT_TEMPLATE = '{current:{num_digits}}'
    TOTAL_TEMPLATE = '{total:{num_digits}}'
    ETA_TEMPLATE = 'ETA: {eta}'  # Estimated Time of Arrival
    SCREW_TEMPLATE = '{screw}'  # rotating ASCII chars
    REFRESH_EVERY = 10

    def __init__(self, console_or_stream=None, **kwargs):
        if isinstance(console_or_stream, Console):
            self.console = console_or_stream
        else:
            self.console = Console(console_or_stream)
        self.template = kwargs.get('template', self.TEMPLATE)
        self.progress_template = \
            kwargs.get('progress_template', self.PROGRESS_TEMPLATE)
        self.progress_brick = kwargs.get('progress_brick', self.PROGRESS_BRICK)
        self.progress_num_bricks = \
            kwargs.get('progress_num_bricks', self.PROGRESS_NUM_BRICKS)
        self.percent_template = \
            kwargs.get('percent_template', self.PERCENT_TEMPLATE)
        self.current_template = \
            kwargs.get('current_template', self.CURRENT_TEMPLATE)
        self.total_template = \
            kwargs.get('total_template', self.TOTAL_TEMPLATE)
        self.eta_template = kwargs.get('eta_template', self.ETA_TEMPLATE)
        self.screw_template = kwargs.get('screw_template', self.SCREW_TEMPLATE)
        self.refresh_every = kwargs.get('refresh_every', self.REFRESH_EVERY)
        self.line_template = self.build_line_template()
        self.start_datetime = None
        self.screw_cycle = itertools.cycle(('|', '/', '-', '\\'))

    def build_line_template(self):
        """Creates line template from other settings"""
        replacements = {
            'progress': self.progress_template,
            'percent': self.percent_template,
            'current': self.current_template,
            'total': self.total_template,
            'eta': self.eta_template,
            'screw': self.screw_template
        }

        def replace(matchobj):
            name = matchobj.group(0)[1:]
            if name not in replacements:
                raise ValueError(
                    u'Unknown ProgressBar.template variable ${0}'.format(name)
                )
            return replacements[name]

        return re.sub('(\$\w+)', replace, self.template)

    def start_timing(self):
        """Manually sets operation's start datetime"""
        self.start_datetime = datetime.datetime.now()

    def end_timing(self):
        """Manually unsets operation's start datetime"""
        self.start_datetime = None

    def refresh_needed(self, current, total):
        """Decides if progress bar should be updated"""
        return current == total or not current % self.refresh_every

    def new_line_needed(self, current, total):
        """Decides if progress bar should write output in new line"""
        return current == total

    def format_message(self, current, total):
        """Creates message to be written on console"""
        if total:
            ratio = float(current) / total
            filled_bricks = int((ratio + 0.05) * self.progress_num_bricks)
            num_digits = int(math.log10(total))
        else:
            ratio = 1.0
            filled_bricks = self.progress_num_bricks
            num_digits = 1
        last_step = total == current
        if last_step:
            eta = datetime.timedelta(0)
        elif ratio and self.start_datetime:
            total_seconds_ = total_seconds(
                datetime.datetime.now() - self.start_datetime
            )
            eta = datetime.timedelta(
                seconds=total_seconds_ / ratio - total_seconds_
            )
        else:
            eta = '?'
        screw = " " if last_step else self.screw_cycle.next()
        return self.line_template.format(
            bricks=self.progress_brick * filled_bricks,
            num_bricks=self.progress_num_bricks,
            ratio=ratio,
            current=current,
            total=total,
            num_digits=num_digits,
            eta=eta,
            screw=screw
        )

    def progress(self, current, total):
        """Updates progress bar with status information"""
        if self.refresh_needed(current, total):
            message = self.format_message(current, total)
            if self.new_line_needed(current, total):
                self.console.new_line(message)
            else:
                self.console.same_line(message)


# Shortcut methods using thread-local ProgressBar and Console instance


def show_progress(current, total=None):
    """Updates progress bar with status information"""
    is_start = current in (0, 1)
    if not hasattr(_thread_locals, 'progress_bar') or is_start:
        _thread_locals.progress_bar = ProgressBar()
    progress_bar = _thread_locals.progress_bar
    if is_start:
        progress_bar.start_timing()
    progress_bar.progress(current, total)


def show_info(message):
    """Writes non-formatted NEW line of text into output stream"""
    if not hasattr(_thread_locals, 'console'):
        _thread_locals.console = Console()
    console = _thread_locals.console
    console.info(message)



# -*- coding: utf-8 -*-

from django.template import TemplateDoesNotExist
from django.template.loaders.base import Loader
from django.core.cache import cache

from powerpages.models import Page
from powerpages import cachekeys


class WebsiteLoader(Loader):
    """
    A custom template loader to load Page templates from the database.
    Inspired by https://github.com/jezdez/django-dbtemplates/
    """

    def load_template_source(self, template_name, template_dirs=None):
        """Load templates from powerpages.Page model instances.
        Works only with templates named:
        page/<page_pk>"""
        try:
            namespace, page_pk = template_name.split('/')
        except ValueError:
            pass
        else:
            if namespace == 'page':
                cachekey = cachekeys.template_source(page_pk)
                display_name = "page:%s" % page_pk
                source = cache.get(cachekey)
                if source is None:
                    try:
                        page = Page.objects.get(pk=page_pk)
                    except Page.DoesNotExist:
                        pass
                    else:
                        page_processor = page.get_page_processor()
                        source = page_processor.get_template_source()
                        cache.set(cachekey, source)
                        display_name = "page:%s" % page_pk
                        return source, display_name
                else:
                    return source, display_name
        raise TemplateDoesNotExist(template_name)



# -*- coding: utf-8 -*-

from django.forms.fields import ChoiceField


class BaseRegistryItemChoiceField(ChoiceField):
    """
    Form field intended to display keys from given class registry
    as it's choices
    """
    # Required configuration:
    registry_instance = None

    def __init__(self, choices=(), *args, **kwargs):
        kwargs.pop('max_length', None)
        # pass choices as callable to defer evaluation:
        choices = self.registry_instance.as_choices
        super(BaseRegistryItemChoiceField, self).__init__(
            choices, *args, **kwargs
        )



# -*- coding: utf-8 -*-

from powerpages.utils.class_registry.registry import ClassRegistry
from powerpages.utils.class_registry import shortcuts


class PageProcessorRegistry(ClassRegistry):
    """Container for all PageProcessor classes"""
    pass


registry = PageProcessorRegistry()

# Module-level shortcut functions:
register = shortcuts.register_function(registry)
unregister = shortcuts.unregister_function(registry)
as_choices = shortcuts.as_choices_function(registry)
get = shortcuts.get_function(registry)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-09-24 12:46
from __future__ import unicode_literals

from django.db import migrations, models
import powerpages.dbfields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Page',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'url',
                    models.CharField(
                        max_length=1024, null=True, verbose_name='URL'
                    )
                ),
                (
                    'alias',
                    models.CharField(
                        blank=True, db_index=True, help_text=(
                            'Unique human-readable codename of page, '
                            'used in template tags'
                        ), max_length=120, null=True, verbose_name='Alias'
                    )
                ),
                (
                    'title',
                    models.CharField(
                        blank=True, default=b'', max_length=512,
                        verbose_name='Title'
                    )
                ),
                (
                    'description',
                    models.TextField(
                        blank=True, default=b'', verbose_name='Description'
                    )
                ),
                (
                    'keywords',
                    models.TextField(
                        blank=True, default=b'', verbose_name='Keywords'
                    )
                ),
                (
                    'template',
                    models.TextField(
                        blank=True, default=b'', help_text=(
                            "Parent's template is used as the base template "
                            "for current page."
                        ), verbose_name='Template'
                    )
                ),
                (
                    'page_processor',
                    powerpages.dbfields.PageProcessorField(
                        default=b'powerpages.DefaultPageProcessor',
                        help_text='Python program used to render this page.',
                        max_length=128, verbose_name='Page Processor'
                    )
                ),
                (
                    'page_processor_config',
                    powerpages.dbfields.PageProcessorConfigField(
                        blank=True, help_text=(
                            'Advanced page configuration options '
                            'as YAML config.'
                        ), null=True, verbose_name='Page Processor Config'
                    )
                ),
                (
                    'is_dirty',
                    models.BooleanField(default=False, editable=False)
                ),
                (
                    'added_at',
                    models.DateTimeField(
                        auto_now_add=True, verbose_name='Added at'
                    )
                ),
                (
                    'changed_at',
                    models.DateTimeField(
                        auto_now=True, verbose_name='Changed at'
                    )
                ),
            ],
            options={
                'ordering': ('url',),
                'verbose_name': 'Page',
                'verbose_name_plural': 'Pages',
            },
        ),
    ]



# -*- coding: utf-8 -*-

from django.test import TestCase
from django.core.urlresolvers import reverse

from powerpages.models import Page


class PageModelTestCase(TestCase):

    # def __unicode__(self):

    def test_unicode(self):
        page = Page.objects.create(
            url='/test/',
        )
        self.assertEqual(unicode(page), '/test/')

    # def get_absolute_url(self):

    def test_get_absolute_url(self):
        page = Page.objects.create(
            url='/test/',
        )
        self.assertEqual(page.get_absolute_url(), '/test/')

    # def parent_url(self):

    def test_parent_url_second_level(self):
        page = Page.objects.create(
            url='/a/b/',
        )
        self.assertEqual(page.parent_url(), '/a/')

    def test_parent_url_first_level(self):
        page = Page.objects.create(
            url='/a/',
        )
        self.assertEqual(page.parent_url(), '/')

    def test_parent_url_root(self):
        page = Page.objects.create(
            url='/',
        )
        self.assertEqual(page.parent_url(), '')

    # def parent(self):

    def test_parent_direct_existing(self):
        parent = Page.objects.create(
            url='/a/',
        )
        page = Page.objects.create(
            url='/a/b/',
        )
        self.assertEqual(page.parent(), parent)

    def test_parent_direct_missing_1(self):
        Page.objects.create(
            url='/a/',
        )
        page = Page.objects.create(
            url='/a/b/c/',
        )
        self.assertIsNone(page.parent())  # /a/b/ does not exist

    def test_parent_direct_missing_2(self):
        page = Page.objects.create(
            url='/a/b/',
        )
        self.assertIsNone(page.parent())

    def test_parent_of_root(self):
        page = Page.objects.create(
            url='/',
        )
        self.assertIsNone(page.parent())

    # def children(self):

    def test_children_empty(self):
        page = Page.objects.create(
            url='/c/',
        )
        Page.objects.create(
            url='/a/b/',
        )
        self.assertEqual(list(page.children()), [])

    def test_children_available(self):
        page = Page.objects.create(
            url='/a/',
        )
        child_1 = Page.objects.create(
            url='/a/b/',
        )
        child_2 = Page.objects.create(
            url='/a/d/',
        )
        Page.objects.create(
            url='/a/b/c/',  # grandchild
        )
        Page.objects.create(
            url='/x/',
        )
        self.assertEqual(list(page.children()), [child_1, child_2])

    # def descendants(self):

    def test_descendants_empty(self):
        page = Page.objects.create(
            url='/c/',
        )
        Page.objects.create(
            url='/a/b/',
        )
        self.assertEqual(list(page.descendants()), [])

    def test_descendants_available(self):
        page = Page.objects.create(
            url='/a/',
        )
        child_1 = Page.objects.create(
            url='/a/b/',
        )
        child_2 = Page.objects.create(
            url='/a/d/',
        )
        grandchild_1 = Page.objects.create(
            url='/a/b/c/',  # grandchild
        )
        Page.objects.create(
            url='/x/',
        )
        self.assertEqual(
            list(page.descendants()), [child_1, grandchild_1, child_2]
        )

    # def is_accessible(self):

    def test_is_accessible_default_processor(self):
        page = Page.objects.create(
            url='/test/'
        )
        self.assertTrue(page.is_accessible())

    def test_is_accessible_redirect_processor(self):
        page = Page.objects.create(
            url='/test/', page_processor='powerpages.RedirectProcessor'
        )
        self.assertTrue(page.is_accessible())

    def test_is_accessible_not_found_processor(self):
        page = Page.objects.create(
            url='/test/', page_processor='powerpages.NotFoundProcessor'
        )
        self.assertFalse(page.is_accessible())

    # def get_admin_url(self):

    def test_get_admin_url(self):
        page = Page.objects.create(
            url='/test/'
        )
        self.assertEqual(
            page.get_admin_url(),
            reverse('admin:powerpages_page_change', args=[page.pk])
        )



# -*- coding: utf-8 -*-

from django.test import TestCase
from django.test.utils import override_settings
from django.core.cache import cache

from powerpages.models import Page
from powerpages import cachekeys


class PageViewTestCase(TestCase):

    # Default Processor:

    def test_page_view_ok(self):
        Page.objects.create(
            url='/test/',
            template='<h1>Hello world!</h1>'
        )
        response = self.client.get('/test/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<h1>Hello world!</h1>')

    def test_page_view_ok_template_inheritance(self):
        Page.objects.create(
            url='/',
            template="""
            <h1>{% block header %}Root{% endblock %}</h1>
            """
        )
        Page.objects.create(
            url='/a/',
            template="""
            {% block header %}Child{% endblock %}
            """
        )
        Page.objects.create(
            url='/a/b/',
            template="""
            {% block header %}Grand{{ block.super }}{% endblock %}
            """
        )
        root_response = self.client.get('/')
        self.assertEqual(root_response.status_code, 200)
        self.assertContains(root_response, '<h1>Root</h1>')

        child_response = self.client.get('/a/')
        self.assertEqual(child_response.status_code, 200)
        self.assertContains(child_response, '<h1>Child</h1>')

        grandchild_response = self.client.get('/a/b/')
        self.assertEqual(grandchild_response.status_code, 200)
        self.assertContains(grandchild_response, '<h1>GrandChild</h1>')

    def test_page_view_ok_custom_base_template(self):
        root = Page.objects.create(
            url='/',
            template="""
            <h1>{% block header %}Root{% endblock %}</h1>
            """
        )
        Page.objects.create(
            url='/a/',
            template="""
            {% block header %}Child{% endblock %}
            """
        )
        Page.objects.create(
            url='/a/b/',
            template="""
            {% block header %}Grand{{ block.super }}{% endblock %}
            """,
            page_processor_config={
                'base template': 'page/{0}'.format(root.pk)
            }
        )

        grandchild_response = self.client.get('/a/b/')
        self.assertEqual(grandchild_response.status_code, 200)
        self.assertContains(grandchild_response, '<h1>GrandRoot</h1>')

    def test_page_view_ok_page_context(self):
        Page.objects.create(
            url='/',
            template="""<title>{{ website_page.title }}</page>
<meta name="keywords" content="{{ website_page.keywords }}">
<meta name="description" content="{{ website_page.description }}">""",
            title="TITLE",
            description="DESCRIPTION",
            keywords="KEYWORDS"
        )
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            """<title>TITLE</page>
<meta name="keywords" content="KEYWORDS">
<meta name="description" content="DESCRIPTION">"""
        )

    def test_page_view_ok_custom_context_processor(self):
        Page.objects.create(
            url='/',
            template="<h1>Magic Number is {{ magic_number }}!</h1>",
            page_processor_config={
                'context processors': [
                    'powerpages.tests.utils.context_processor'
                ]
            }
        )
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "<h1>Magic Number is 42!</h1>")

    def test_page_view_ok_request_context_processor(self):
        Page.objects.create(
            url='/',
            template="<p>HTTP method was: {{ request.method }}</p>",
        )
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "<p>HTTP method was: GET</p>")

    def test_page_view_ok_tags_available(self):
        Page.objects.create(
            url='/test/',
            template='<a href="{% page_url other-page %}">Other</a>'
        )
        Page.objects.create(
            url='/other/',
            alias='other-page'
        )
        response = self.client.get('/test/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<a href="/other/">Other</a>')

    def test_page_view_ok_custom_tags(self):
        Page.objects.create(
            url='/test/',
            template="<h1>Magic Number is {% get_magic_number %}!</h1>",
            page_processor_config={
                'tag libraries': ['powerpages_test']
            }
        )
        response = self.client.get('/test/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "<h1>Magic Number is 42!</h1>")

    def test_page_view_ok_custom_headers(self):
        Page.objects.create(
            url='/test/',
            page_processor_config={
                'headers': {'X-Magic-Number': '42'}
            }
        )
        response = self.client.get('/test/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['X-Magic-Number'], '42')

    @override_settings(
        CACHES={
            'default': {
                'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
                'LOCATION': 'powerpages-test'
            }
        }
    )
    def test_page_view_ok_with_cache(self):
        page = Page.objects.create(
            url='/test/',
            template='<h1>Hello world!</h1>',
            page_processor_config={'cache': 15}  # cache for 15 seconds
        )
        cache_key = cachekeys.rendered_source_for_lang(page.pk, 'en')
        self.assertNotIn(cache_key, cache)
        response = self.client.get('/test/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '<h1>Hello world!</h1>')
        self.assertIn(cache_key, cache)

    # Not Found Processor:

    def test_page_view_not_found(self):
        Page.objects.create(
            url='/test/',
            template='<h1>Hello world!</h1>',
            page_processor='powerpages.NotFoundProcessor'
        )
        response = self.client.get('/test/')
        self.assertEqual(response.status_code, 404)

    # Redirect Processor:

    def test_page_view_redirect(self):
        Page.objects.create(
            url='/test/',
            template='<h1>Hello world!</h1>',
            page_processor='powerpages.RedirectProcessor',
        )
        response = self.client.get('/test/')
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'], '/')

    def test_page_view_redirect_to_url(self):
        Page.objects.create(
            url='/old/',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={'to url': '/new/'}
        )
        Page.objects.create(
            url='/new/',
        )
        response = self.client.get('/old/')
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'], '/new/')

    def test_page_view_redirect_to_alias(self):
        Page.objects.create(
            url='/old/',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={'to alias': 'new_page'}
        )
        Page.objects.create(
            url='/new/',
            alias='new_page'
        )
        response = self.client.get('/old/')
        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'], '/new/')

    def test_page_view_redirect_302(self):
        Page.objects.create(
            url='/old/',
            page_processor='powerpages.RedirectProcessor',
            page_processor_config={'to url': '/new/', 'permanent': False}
        )
        Page.objects.create(
            url='/new/',
            alias='new_page'
        )
        response = self.client.get('/old/')
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/new/')



# -*- coding=utf-8 -*-

import copy
from importlib import import_module


default_app_config = 'powerpages.apps.PowerPagesConfig'


def autodiscover():
    """
    Auto-discover INSTALLED_APPS modules:
    * page_processors.py to fill out the default PageProcessorRegistry,
    * sitemap.py to populate sitemap_config.sitemaps.
    Inspired by django.contrib.admin.autodiscover.
    """

    from django.conf import settings
    from django.utils.module_loading import module_has_submodule
    from powerpages import page_processor_registry

    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        # page processors:
        try:
            before_import_registry = copy.copy(
                page_processor_registry.registry.registry
            )
            import_module('%s.page_processors' % app)
        except:
            page_processor_registry.registry.registry = before_import_registry
            if module_has_submodule(mod, 'page_processors'):
                raise
        # sitemaps:
        try:
            import_module('%s.sitemap' % app)
        except:
            if module_has_submodule(mod, 'sitemap'):
                raise



# -*- coding: utf-8 -*-

from django.core.cache import cache
from django.dispatch import receiver
from django.db import models

from powerpages.settings import app_settings
from powerpages.utils.attribute_cache import cache_result_on
from powerpages.dbfields import (
    PageProcessorField, PageProcessorConfigField
)
from powerpages import cachekeys


# Helpers:

class PageURLCache(object):
    """Helps to retrieve Page URL by alias using cache"""

    @staticmethod
    def refresh():
        """
        Build dictionary with relation alias <-> url for regular pages
        so they can be faster accessible from other places
        """
        alias_to_url = dict(
            Page.objects.filter(alias__isnull=False)
                        .values_list('alias', 'url')
        )
        cache.set(
            cachekeys.URL_LIST_CACHE, alias_to_url, app_settings.CACHE_SECONDS
        )
        return alias_to_url

    @staticmethod
    def url_by_alias(page_alias):
        """
        Try to read page url from cache. If it's missing then try to find
        matching page that could be missing in cache. If page is found
        then refresh all url list since it's too old.
        If no matching is found then return None so we can throw any
        exception we want in other places
        """
        if page_alias:
            url_to_alias = cache.get(cachekeys.URL_LIST_CACHE)
            if url_to_alias is None:
                url_to_alias = PageURLCache.refresh()
            url = url_to_alias.get(page_alias)
        else:
            url = None
        return url


# Models:

class Page(models.Model):
    """Webpages and templates."""

    class Meta:
        verbose_name = u"Page"
        verbose_name_plural = u"Pages"
        ordering = ('url',)

    # Identity:
    url = models.CharField(
        verbose_name=u"URL", max_length=1024, null=True
    )
    alias = models.CharField(
        verbose_name=u"Alias", max_length=120, db_index=True,
        null=True, blank=True,
        help_text=u"Unique human-readable codename of page, "
                  u"used in template tags"
    )
    # Meta tags:
    title = models.CharField(
        max_length=512, verbose_name=u"Title", blank=True, default=""
    )
    description = models.TextField(
        verbose_name=u"Description", blank=True, default=""
    )
    keywords = models.TextField(
        verbose_name=u"Keywords", blank=True, default=""
    )
    # Template:
    template = models.TextField(
        verbose_name=u"Template",
        blank=True, default='',
        help_text=u"Parent's template is used as the base template for "
                  u"current page."
    )
    # Page processing
    page_processor = PageProcessorField(
        verbose_name=u"Page Processor", max_length=128,
        default='powerpages.DefaultPageProcessor',
        help_text=u"Python program used to render this page."
    )
    page_processor_config = PageProcessorConfigField(
        verbose_name=u"Page Processor Config", null=True, blank=True,
        help_text=u"Advanced page configuration options as YAML config."
    )
    # Indicates objects saved in Admin:
    is_dirty = models.BooleanField(default=False, editable=False)
    # Change info fields:
    added_at = models.DateTimeField(
        verbose_name=u"Added at", auto_now_add=True
    )
    changed_at = models.DateTimeField(
        verbose_name=u"Changed at", auto_now=True
    )

    def __unicode__(self):
        """URL"""
        return self.url

    def get_absolute_url(self):
        """URL"""
        return self.url

    def parent_url(self):
        """Generates parent Page URL"""
        url = self.url
        if url.endswith('/'):
            url = url[:-1]
        parent_url = url.rsplit('/', 1)[0] + '/'
        return parent_url if self.url != parent_url else ''

    @cache_result_on('_parent')
    def parent(self):
        """Parent Page based on URL"""
        parent_url = self.parent_url()
        if parent_url:
            parent_page = Page.objects.filter(url=parent_url).first()
        else:
            parent_page = None
        return parent_page

    def children(self):
        """Direct descendant Pages based on URL"""
        url_regex = '^{url}[^/]+/?$'.format(url=self.url)
        return Page.objects.filter(url__regex=url_regex)

    def descendants(self):
        """Direct and indirect descendant Pages based on URL"""
        url = self.url
        return Page.objects.filter(url__startswith=url).exclude(url=url)

    def is_accessible(self):
        """Determines if Page can be accessed on URL"""
        page_processor = self.get_page_processor()
        return page_processor and page_processor.is_accessible()

    @models.permalink
    def get_admin_url(self):
        """URL of Page edition view in Admin"""
        return 'admin:powerpages_page_change', [self.pk]


# Signal Receivers:


@receiver(models.signals.post_save, sender=Page)
def page_changed(sender, **kwargs):
    """
    post_save receiver for Page model:
    * clears Page-related cache keys,
    * refresh mappings: alias <-> page real url,
    * clears cache keys related to PageChanges.
    """
    page = kwargs['instance']
    cache_key = cachekeys.template_source(page.pk)
    cache.delete(cache_key)
    PageURLCache.refresh()



# -*- coding: utf-8 -*-

import difflib
from cStringIO import StringIO

from django.utils.html import escape


class Diff(object):
    """Text diff"""

    CHANGE_CLASSES = {
        '+': "diff-added",
        '-': "diff-removed",
        ' ': None
    }

    def __init__(self):
        self.left_side = None
        self.right_side = None

    def left(self, content, created_at):
        """Creates DiffSide objects and sets it as left side"""
        self.left_side = DiffSide("Old", content, created_at)

    def right(self, content, created_at):
        """Creates DiffSide objects and sets it as right side"""
        self.right_side = DiffSide("New", content, created_at)

    def html_diff(self):
        """Creates unified diff as HTML"""
        if not (self.left_side and self.right_side):
            raise RuntimeError(
                "Both sides of diff must be set before calling html_diff"
            )
        diff_gen = difflib.unified_diff(
            self.left_side.content, self.right_side.content,
            self.left_side.name, self.right_side.name,
            self.left_side.datetime_as_string(),
            self.right_side.datetime_as_string(),
            n=50
        )
        output = StringIO()
        for line in diff_gen:
            if len(line) == 2:
                change, char = line
                change_class = self.CHANGE_CLASSES[change]
                if change_class:
                    marked_line = u'<span class="%s">%s</span>' % (
                        change_class, escape(char)
                    )
                else:
                    marked_line = escape(char)
            else:
                marked_line = u'\n<span class="diff-info">%s</span>\n' % (
                    escape(line[:-2]),
                )
            output.write(marked_line.encode('utf-8'))
        return output.getvalue()


class DiffSide(object):
    """One side of text diff"""

    def __init__(self, name, content, created_at):
        self.name = name
        self.content = content
        self.created_at = created_at

    def datetime_as_string(self):
        """Converts creation datetime to ISO 8601 string."""
        return self.created_at.strftime("%Y-%m-%d %H:%M:%S")



# -*- coding: utf-8 -*-

from django import forms


class SourceCodeEditor(forms.Textarea):
    """Textarea being source code editor"""

    class Media:
        js = ('powerpages/js/ace/ace.js', 'powerpages/js/bind_ace_editor.js')

    HANDLER_CLASS = 'handle-ace-editor'

    def __init__(self, attrs=None):
        super(SourceCodeEditor, self).__init__(attrs)
        # Add handler class to widget's classes
        classes = [
            cls for cls in self.attrs.get('class', '').split(' ') if cls
        ]
        if self.HANDLER_CLASS not in classes:
            classes.insert(0, self.HANDLER_CLASS)
        self.attrs['class'] = ' '.join(classes)



# -*- coding: utf-8 -*-


class ClassRegistry(object):
    """Collects classes and allows to list them."""

    class RegistryError(Exception):
        """Base class for all registry errors"""
        pass

    class AlreadyRegistered(RegistryError):
        """Class is already registered"""
        pass

    class NotRegistered(RegistryError):
        """Class is not registered"""
        pass

    def __init__(self):
        self.registry = {}

    def get_key_from_class(self, cls):
        """Prepare registry key for given class"""
        return "%s.%s" % (cls.__module__.split('.')[-2], cls.__name__)

    def register(self, cls):
        """Registers new class"""
        key = self.get_key_from_class(cls)
        if key in self.registry:
            raise self.AlreadyRegistered(key)
        self.registry[key] = cls

    def unregister(self, cls):
        """Unregisters the class"""
        key = self.get_key_from_class(cls)
        if key not in self.registry:
            raise self.NotRegistered(key)
        self.registry.pop(key)

    def as_choices(self):
        """Get choice list suitable for form's ChoiceField"""
        return sorted(
            (
                (k, self.get_item_label(k, v))
                for (k, v) in self.registry.items()
            ),
            key=lambda t: t[1]
        )

    def get(self, key):
        """Retrieves class for given key"""
        if key not in self.registry:
            raise self.NotRegistered(key)
        return self.registry[key]

    def get_item_label(self, key, value):
        """Provides label for item with fallback to raw key"""
        return getattr(value, 'verbose_name', key)

    def configure_item(self, item_class, model_instance, config):
        """
        Configure registry item class per model instance and config.
        Required to use ClassRegistry together with BaseRegistryItemField.
        Default implementation creates instance of class.
        """
        return item_class(model_instance, config)



# -*- coding: utf-8 -*-

import yaml

from django import forms
from django.db import models
from django.core.serializers.pyyaml import DjangoSafeDumper


class BaseRegistryItemField(models.CharField):
    """
    Model field intended to store keys from given class registry
    """
    # Required configuration:
    registry_instance = None
    item_accessor_method_name = None
    form_field_class = None
    # Item configuration (optional):
    config_accessor_method_name = None

    def formfield(self, form_class=None, **kwargs):
        if form_class is None:
            form_class = self.form_field_class
        kwargs['form_class'] = form_class
        kwargs['widget'] = forms.Select
        return super(BaseRegistryItemField, self).formfield(**kwargs)

    def contribute_to_class(self, cls, name, *args, **kwargs):
        """Apply get_page_processor method to model class"""

        def get_registry_item(model_instance):
            key = getattr(model_instance, name)
            if key:
                item_class = self.registry_instance.get(key)
                if self.item_accessor_method_name:
                    item_config = getattr(
                        model_instance, self.config_accessor_method_name,
                        lambda: {}
                    )()
                else:
                    item_config = None
                # Class registry is responsible for configuration
                # (eg. instantiation) of items
                return self.registry_instance.configure_item(
                    item_class, model_instance, item_config
                )

        super(
            BaseRegistryItemField, self
        ).contribute_to_class(cls, name, *args, **kwargs)
        setattr(cls, self.item_accessor_method_name, get_registry_item)


class BaseRegistryItemConfigField(models.TextField):
    """
    Model field for storing configuration in database as YAML documents.
    Based on:
    https://github.com/datadesk/django-yamlfield
    """
    # Required configuration:
    config_accessor_method_name = None

    def from_db_value(self, value, expression, connection, context):
        """
        Convert our YAML string to a Python object
        after we load it from the DB.
        """
        if value == "":
            return None
        try:
            if isinstance(value, basestring):
                return yaml.load(value)
        except ValueError:
            pass
        return value

    def get_db_prep_save(self, value, connection):
        """
        Convert our Python object to a string of YAML before we save.
        """
        if not value or value == "":
            return ""
        if isinstance(value, (dict, list)):
            value = yaml.dump(
                value, Dumper=DjangoSafeDumper, default_flow_style=False
            )
        return super(BaseRegistryItemConfigField, self).get_db_prep_save(
            value, connection=connection
        )

    def value_from_object(self, obj):
        """
        Returns the value of this field in the given model instance.

        We need to override this so that the YAML comes out properly formatted
        in the admin widget.
        """
        value = getattr(obj, self.attname)
        if not value or value == "":
            return value
        return yaml.dump(
            value, Dumper=DjangoSafeDumper, default_flow_style=False
        )

    def contribute_to_class(self, cls, name, *args, **kwargs):
        """Apply get_page_processor_config method to model class"""

        def get_registry_item_config(model_instance):
            return getattr(model_instance, name) or {}

        super(
            BaseRegistryItemConfigField, self
        ).contribute_to_class(cls, name, *args, **kwargs)
        setattr(
            cls, self.config_accessor_method_name, get_registry_item_config
        )



# -*- coding: utf-8 -*-

from django.core.validators import ValidationError


def default_converter(value):
    return value


class UndefinedVariable(Exception):
    """Attempt to use variable missing from the specification."""
    pass


class Config(object):
    """Container for variables' definitions and applied values"""

    def __init__(self, variables, data):
        self.variable_by_name = {}
        for variable in variables:
            self.variable_by_name[variable.name] = variable
        self.data = data

    def get(self, name):
        """Retrieves value from config"""
        if name not in self.variable_by_name:
            raise UndefinedVariable(name)
        variable = self.variable_by_name[name]
        return variable.get_value(self.data)

    def validate(self):
        """
        Checks the data against validation settings for each variable.
        """
        errors = []
        for name, variable in self.variable_by_name.items():
            try:
                variable.validate_value(self.data)
            except ValidationError as e:
                for message in e.messages:
                    errors.append(u'{0}: {1}'.format(name, message))
        if errors:
            raise ValidationError(errors)


class ConfigVariable(object):
    """Specification of config variable"""

    default_error_messages = {
        'required': u'This variable is required.',
        'invalid': u'Enter a valid value.',
        'unable_to_process':
        u'Unable to process variable (%(exc_type)s: %(exc_message)s).',
        'invalid_choice': u'%(value)s is not one of the available choices.',
    }

    def __init__(self, name, converter=None, default=None, choices=None,
                 multiple=False, required=False, validators=None,
                 error_messages=None, help_text=''):
        if converter and choices:
            raise ValueError(
                'Options `converter` and `choices` are exclusive.'
            )
        self.name = name
        self.converter = converter if converter else default_converter
        self.default = default
        self.choices = choices
        self.multiple = multiple
        self.required = required
        self.help_text = help_text
        self.validators = validators or []
        # Error message gathering:
        messages = {}
        for c in reversed(self.__class__.__mro__):
            messages.update(getattr(c, 'default_error_messages', {}))
        messages.update(error_messages or {})
        self.error_messages = messages

    def get_default_value(self):
        """Retrieves default value with optional validation"""
        value = self.default
        if callable(value):  # may be callable
            value = value()
        return value

    def get_single_value(self, raw_config_value):
        """Retrieves single value"""
        if self.choices:
            # Choices applied:
            # Map the value using choices, with fallback to the default
            # when value is not in available choices
            reversed_choice_dict = dict((v, k) for (k, v) in self.choices)
            try:
                value = reversed_choice_dict[raw_config_value]
            except KeyError:
                raise ValidationError(
                    self.error_messages['invalid_choice'] % {
                        'value': raw_config_value
                    }
                )
        else:
            # Converter applied:
            # Map the value using converter,
            # with fallback to the default when conversion fails
            try:
                value = self.converter(raw_config_value)
            except Exception as e:
                raise ValidationError(
                    self.error_messages['unable_to_process'] % {
                        'exc_type': e.__class__.__name__,
                        'exc_message': e.message
                    }
                )
        if callable(value):  # may be callable
            value = value()
        return value

    def get_multiple_values(self, raw_config_values):
        """Retrieves multiple values"""
        try:
            iter(raw_config_values)
        except TypeError as e:
            raise ValidationError(
                self.error_messages['unable_to_process'] % {
                    'exc_type': e.__class__.__name__,
                    'exc_message': e.message
                }
            )
        return [
            self.get_single_value(raw_config_value)
            for raw_config_value in raw_config_values
        ]

    def get_value(self, data):
        """Retrieves value from data with fallback to defaults"""
        # Retrieve raw config value, with fallback to default
        # when value is not defined
        try:
            raw_config_value = data[self.name]
        except KeyError:
            value = self.get_default_value()
        else:
            try:
                if self.multiple:
                    value = self.get_multiple_values(raw_config_value)
                else:
                    value = self.get_single_value(raw_config_value)
            except ValidationError:
                value = self.get_default_value()
        return value

    def validate_value(self, data):
        """Checks the data against validation settings this variable."""
        value = self.run_basic_validation(data)
        self.run_validators(value)

    def run_basic_validation(self, data):
        """
        Validates presence of variable and conversion.
        Returns converted value on success.
        """
        # Check presence of value
        try:
            raw_config_value = data[self.name]
        except KeyError:
            if self.required:
                raise ValidationError(self.error_messages['required'])
            value = self.get_default_value()
        else:
            # Validation errors are passed:
            if self.multiple:
                value = self.get_multiple_values(raw_config_value)
            else:
                value = self.get_single_value(raw_config_value)
        return value

    def run_validators(self, value):
        """
        Checks value retrieved by run_basic_validation()
        against all validators.
        Copied from `django.forms.fields.Field.run_validators`.
        """
        errors = []
        for validator in self.validators:
            try:
                validator(value)
            except ValidationError as e:
                if hasattr(e, 'code') and e.code in self.error_messages:
                    message = self.error_messages[e.code]
                    if e.params:
                        message = message % e.params
                    errors.append(message)
                else:
                    errors.extend(e.messages)
        if errors:
            raise ValidationError(errors)



# -*- coding: utf-8 -*-

import json
import hashlib


URL_LIST_CACHE = "website:url_list"
SITEMAP_CONTENT = "website:sitemap"


def get_cache_name(prefix, name):
    """
    Cache name constructor. Uses the same methods as django cache system
    Examples:
    *) prefix=profile.cache, name=<requestuser.id>
    *) prefix=template.cache.sidebar, name=<requestuser.id>
    """
    return u'{0}.{1}'.format(
        prefix, hashlib.md5(unicode(name).encode('utf-8')).hexdigest()
    )


def template_source(page_pk):
    """Create cache key for page template"""
    return 'website:template:%s' % page_pk


def rendered_source_for_user(page_pk, user_id):
    """Create cache key for rendered page source based on current user"""
    return 'website:rendered_source_user:%s:%s' % (page_pk, user_id)


def rendered_source_for_lang(page_pk, lang):
    """Create cache key for rendered page source based on current language"""
    return 'website:rendered_source_lang:%s:%s' % (page_pk, lang)


def url_cache(name, *args, **kwargs):
    """
    Creates cache key for url of CMS page or standard Django URL
    based on hashed serialized name with optional *args and **kwargs
    """
    serialized_url = json.dumps([name, args, kwargs], sort_keys=True)
    return get_cache_name("website:urls", serialized_url)



# -*- coding: utf-8 -*-

from django.conf import settings
from django.test.signals import setting_changed


DEFAULTS = {
    'CACHE_SECONDS': 60 * 60,  # 1 hour
    'SYNC_DIRECTORY': None,
    'TAG_LIBRARIES': (),
    'SITEMAP_PROTOCOL': None,
    'SITEMAP_DOMAIN': None,
    'SITEMAP_DEFAULT_CHANGEFREQ': None,
    'SITEMAP_DEFAULT_PRIORITY': None,
}


class AppSettings(object):
    """
    A settings object that allows app settings to be accessed by properties.
    Inspired by solution from Django REST Framework.
    """

    def __init__(self, user_settings=None, defaults=None):
        self.user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.cache = {}

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError('Invalid app setting "{0}"'.format(attr))
        try:
            value = self.cache[attr]
        except KeyError:
            if self.user_settings is None:
                self.user_settings = getattr(settings, 'POWER_PAGES', {})
            try:
                value = self.user_settings[attr]
            except KeyError:
                value = self.defaults[attr]
            self.cache[attr] = value
        return value


app_settings = AppSettings()


def reload_app_settings(*args, **kwargs):
    if kwargs['setting'] == 'POWER_PAGES':
        app_settings.user_settings = kwargs['value']
        app_settings.cache = {}


setting_changed.connect(reload_app_settings)



# -*- coding: utf-8 -*-

from powerpages.utils.class_registry.dbfields import (
    BaseRegistryItemField, BaseRegistryItemConfigField
)
from powerpages.formfields import PageProcessorChoiceField
from powerpages import page_processor_registry


class PageProcessorField(BaseRegistryItemField):
    """
    Model field intended to store keys from page_processor_registry.registry
    """
    # Required configuration:
    registry_instance = page_processor_registry.registry
    item_accessor_method_name = 'get_page_processor'
    form_field_class = PageProcessorChoiceField
    # Item configuration (optional):
    config_accessor_method_name = 'get_page_processor_config'


class PageProcessorConfigField(BaseRegistryItemConfigField):
    """
    Model field for storing configuration of PageProcessors.
    """
    # Required configuration:
    config_accessor_method_name = 'get_page_processor_config'



# -*- coding: utf-8 -*-

from .website_base import BaseDumpLoadCommand
from powerpages.sync import WebsiteLoadOperation


class Command(BaseDumpLoadCommand):
    """LOADS website Pages FROM structure of directories and file"""

    operation_class = WebsiteLoadOperation



# -*- coding=utf-8 -*-

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class PowerPagesConfig(AppConfig):
    name = 'powerpages'
    verbose_name = _('CMS')

    def ready(self):
        """Register website data and add default templatetags"""
        # Autodiscover website PageProcessors
        from powerpages import autodiscover
        autodiscover()



# -*- coding: utf-8 -*-

from django.conf.urls import url
from powerpages import views


urlpatterns = [
    url('^powerpages-admin/switch-edit-mode/$', views.admin_switch_edit_mode,
        name='switch_edit_mode'),
    url(r'^sitemap\.xml', views.sitemap, name='sitemap'),
    url('^(?P<path>[\S\s]*)$', views.page, name='page')
]



# -*- coding: utf-8 -*-

import os

from setuptools import setup, find_packages
from distutils.core import Command


here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read()


class DjangoCommand(Command):

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from django.conf import settings
        settings.configure(
            INSTALLED_APPS=(
                'django.contrib.auth',
                'django.contrib.admin',
                'django.contrib.contenttypes',
                'django.contrib.sessions',
                'django.contrib.messages',
                'django.contrib.staticfiles',
                'powerpages',
            ),
            SECRET_KEY='AAA',
            DATABASES={
                'default': {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3'
                }
            },
            # TODO: MIDDLEWARE for 1.10
            MIDDLEWARE_CLASSES=(
                'django.middleware.security.SecurityMiddleware',
                'django.contrib.sessions.middleware.SessionMiddleware',
                'django.middleware.common.CommonMiddleware',
                'django.middleware.csrf.CsrfViewMiddleware',
                'django.contrib.auth.middleware.AuthenticationMiddleware',
                'django.contrib.messages.middleware.MessageMiddleware',
                'django.middleware.clickjacking.XFrameOptionsMiddleware',
                'django.middleware.locale.LocaleMiddleware',
            ),
            LANGUAGE_CODE='en-us',
            ROOT_URLCONF='powerpages.tests.urls',
            TEMPLATES=[
                {
                    'BACKEND':
                    'django.template.backends.django.DjangoTemplates',
                    'OPTIONS': {
                        'loaders': [
                            'django.template.loaders.filesystem.Loader',
                            'django.template.loaders.app_directories.Loader',
                            # Database template loader for powerpages app
                            'powerpages.loader.WebsiteLoader',
                        ],
                        'context_processors': [
                            'django.template.context_processors.debug',
                            'django.template.context_processors.request',
                            'django.contrib.auth.context_processors.auth',
                            'django.contrib.messages.context_processors.messages',
                        ]
                    },
                },
            ]
        )
        import django
        django.setup()
        self._run()

    def _run(self):
        raise NotImplementedError


class Test(DjangoCommand):

    def _run(self):
        from django.core.management import call_command
        call_command('test', 'powerpages')


class MakeMigrations(DjangoCommand):

    def _run(self):
        from django.core.management import call_command
        call_command('makemigrations', 'powerpages')


setup(
    name='django-powerpages',
    version='0.0.2',
    description=(
        'Developer-friendly, simple CMS for Django, "flatpages on steroids".'
    ),
    long_description=long_description,
    url='https://github.com/Open-E-WEB/django-powerpages',
    author='Open-E, Inc.',
    author_email='',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django :: 1.9',
        'Topic :: Internet :: WWW/HTTP',
    ],
    keywords='django cms web',
    install_requires=[
        'PyYAML==3.11',
    ],
    packages=find_packages(),
    include_package_data=True,
    cmdclass={'test': Test, 'makemigrations': MakeMigrations},
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import sys
import json
import codecs
import itertools
import difflib
import collections
import subprocess

from powerpages.utils.console import Console
from powerpages.models import Page
from powerpages.settings import app_settings


INDEX_FILE_NAME = '_index_.page'

FILE_DELIMITER = '## TEMPLATE SOURCE: ##'
FILE_DELIMITER_LINE = '\n{0}\n'.format(FILE_DELIMITER)


def url_to_path(url, has_children):
    """Generates file system path for given URL"""
    url_pieces = [slug for slug in url.split('/') if slug]
    if url == '/':
        rel_path = INDEX_FILE_NAME
    elif url_pieces and has_children:
        rel_path = os.sep.join(url_pieces + [INDEX_FILE_NAME])
    else:
        file_name = url_pieces.pop()
        if '.' not in file_name:
            file_name = '{0}.page'.format(file_name)
        rel_path = os.sep.join(url_pieces + [file_name])
    return rel_path


def path_to_url(path):
    """Generates URL for given relative file path"""
    url_pieces = [slug for slug in path.split('/') if slug]
    file_name = url_pieces.pop()
    if file_name != INDEX_FILE_NAME:
        if file_name.endswith('.page'):
            file_name, file_ext = os.path.splitext(file_name)
        url_pieces.append(file_name)
        if '.' not in file_name:
            url_pieces.append('')  # to generate trailing slash
    else:
        url_pieces.append('')  # to generate trailing slash
    return '/' + '/'.join(url_pieces)


def generate_diff(current, coming):
    """Generates diff of changes"""
    return u'\n'.join(
        difflib.unified_diff(
            current.splitlines(), coming.splitlines(),
            u"Current content", u"Coming changes"
        )
    )


def normalize_text(s):
    """Normalizes content of a text field"""
    return (s or '').strip().replace('\r', '')


def normalize_line(s):
    """Normalizes content of a field that should be a single line, eg. title"""
    return normalize_text(s).replace('\n', '')


def normalize_template(s):
    """Normalizes content of a template"""
    normalized = normalize_text(s)
    if normalized:
        normalized += '\n'
    return normalized


def normalize_page_fields(data):
    cleaned_data = {
        'alias': data.get('alias'),
        'page_processor_config': data.get('page_processor_config'),
        'template': normalize_template(data.get('template'))
    }
    for f in ('title', 'page_processor'):
        cleaned_data[f] = normalize_text(data.get(f))
    for f in ('description', 'keywords'):
        cleaned_data[f] = normalize_text(data.get(f))
    return cleaned_data


class SyncStatus(object):
    ADDED = 'A'
    MODIFIED = 'M'
    NO_CHANGES = '.'
    DELETED = 'D'
    # Additional status:
    SKIPPED = 's'
    FORCED = '!'
    # Descriptions:
    DESCRIPTIONS = {
        ADDED: u"Added",
        MODIFIED: u"Modified",
        NO_CHANGES: u"No changes",
        DELETED: u"Deleted",
        SKIPPED: u"(skipped)",
        FORCED: u"(forced)",
    }

    @classmethod
    def describe(cls, status):
        """Descriptive version of status"""
        return u" ".join(cls.DESCRIPTIONS[s] for s in status)


class PageFileDumper(object):
    """Class responsible for DUMPING Page instance data INTO file"""

    def __init__(self, page):
        self.page = page

    def relative_path(self):
        """Relative path to file"""
        return url_to_path(self.page.url, self.page.children().exists())

    def absolute_path(self):
        """Absolute path to file"""
        return os.path.join(app_settings.SYNC_DIRECTORY, self.relative_path())

    def page_fields(self):
        """Dictionary of Page fields"""
        page = self.page
        page_fields = {
            'alias': page.alias,
            'title': page.title,
            'description': page.description,
            'keywords': page.keywords,
            'template': page.template,
            'page_processor': page.page_processor,
            'page_processor_config': page.page_processor_config,
        }
        return normalize_page_fields(page_fields)

    @staticmethod
    def dump(page_fields):
        """Converts dict of Page fields into file contents"""
        template_source = page_fields.pop('template') or ''
        page_metadata = json.dumps(
            collections.OrderedDict(
                sorted((k, v) for k, v in page_fields.items())
            ),
            indent=2
        )
        return FILE_DELIMITER_LINE.join(
            (page_metadata, template_source)
        )

    def file_contents(self):
        """Contents of .page file"""
        return self.dump(self.page_fields())

    def file_exists(self):
        """File exists?"""
        return os.path.exists(self.absolute_path())

    def status(self):
        """Synchronization status determined before actual synchronization"""
        if not self.file_exists():
            status = SyncStatus.ADDED
        elif (
            # previous data
            FilePageLoader(self.relative_path()).page_fields() !=
            # is different from current data
            self.page_fields()
        ):
            status = SyncStatus.MODIFIED
        else:
            status = SyncStatus.NO_CHANGES
        return status

    def save(self):
        """Saves Page data into file"""
        # Side effect: remove dirty flag:
        Page.objects.filter(pk=self.page.pk).update(is_dirty=False)
        # Ensure directory:
        dir_path = os.path.dirname(self.absolute_path())
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        # Save file:
        with codecs.open(self.absolute_path(), 'w', encoding='utf-8') as f:
            f.write(self.file_contents())

    def diff(self):
        """Generates diff of changes (normalized)"""
        normalized_current = PageFileDumper.dump(
            FilePageLoader(self.relative_path()).page_fields()
        )
        normalized_coming = self.file_contents()
        return generate_diff(normalized_current, normalized_coming)


class FilePageLoader(object):
    """Class responsible for LOADING Page instance data FROM file"""

    def __init__(self, path):
        self.relative_path = path

    def absolute_path(self):
        """Absolute path to file"""
        return os.path.join(app_settings.SYNC_DIRECTORY, self.relative_path)

    def url(self):
        """URL based on relative file path"""
        return path_to_url(self.relative_path)

    @staticmethod
    def load(file_contents):
        """Converts file contents into dict of Page fields"""
        page_metadata, template_source = \
            file_contents.split(FILE_DELIMITER_LINE, 1)
        page_fields = json.loads(page_metadata)
        page_fields['template'] = template_source
        return page_fields

    def page_fields(self):
        """Dictionary of Page fields"""
        try:
            page_fields = self.load(self.file_contents())
        except ValueError:
            raise RuntimeError(
                u'Bad .page file: {0}'.format(self.relative_path)
            )
        return normalize_page_fields(page_fields)

    def file_contents(self):
        """Contents of .page file"""
        return codecs.open(
            self.absolute_path(), 'r', encoding='utf-8'
        ).read()

    def page(self):
        """Corresponding Page instance"""
        return Page.objects.filter(url=self.url()).first()

    def status(self):
        """Synchronization status determined before actual synchronization"""
        page = self.page()
        if not page:
            status = SyncStatus.ADDED
        elif (
            # Page data in database
            PageFileDumper(page).page_fields() !=
            # is different from file data
            self.page_fields()
        ):
            status = SyncStatus.MODIFIED
        else:
            status = SyncStatus.NO_CHANGES
        return status

    def save(self):
        """Saves file data into Page"""
        url = self.url()
        page_fields = self.page_fields()
        page = Page.objects.filter(url=url).first()
        if not page:
            page = Page(url=url)
        for name, value in page_fields.items():
            setattr(page, name, value)
        page.is_dirty = False  # Remove dirty flag
        page.save()

    def diff(self):
        """Generates diff of changes (normalized)"""
        normalized_current = PageFileDumper(self.page()).file_contents()
        normalized_coming = PageFileDumper.dump(self.page_fields())
        return generate_diff(normalized_current, normalized_coming)


class BaseSyncOperation(object):
    """Base class for website dump / load operations"""

    status_colors = {
        SyncStatus.DELETED: 'MAGENTA',
        SyncStatus.ADDED: 'GREEN',
        SyncStatus.MODIFIED: 'CYAN',
    }

    def __init__(self, root_url=None, error_class=None,
                 stdout=None, stderr=None, get_input=None, **options):
        self.root_url = root_url or '/'
        self.error_class = error_class or RuntimeError
        self.stdout = stdout or sys.stdout
        self.stderr = stderr or sys.stderr
        self.dry_run = options['dry_run']
        self.no_interactive = options['no_interactive']
        self.quiet = options['quiet']
        self.force = options['force']
        self.git_add = options['git_add']
        self.no_color = options.get('no_color')
        self.console = Console(self.stdout)
        self.get_input = get_input or raw_input

    def status_fg_color(self, status):
        """Determines fg color for status"""
        if self.no_color:
            fg_color = None
        else:
            try:
                base_status, extra_status = status
            except ValueError:
                base_status, extra_status = status, None
            if extra_status == SyncStatus.SKIPPED:
                fg_color = None
            else:
                fg_color = self.status_colors.get(base_status, None)
        return fg_color

    def log(self, message):
        """Writes informative line of text to stdout"""
        if not self.quiet:
            self.console.new_line(message)

    def log_status(self, status, item):
        """Writes informative line about status of given item (in color!)"""
        if not self.quiet:
            message = u"{0} {1}".format(status, item)
            self.console.new_line(
                message, fg_color=self.status_fg_color(status)
            )

    def summary(self, summary_dict):
        """Writes summary line of text to stdout"""
        self.console.new_line(u"SUMMARY:")
        if summary_dict:
            for status, occurrences in summary_dict.items():
                message = u"\t{0} [{1}] = {2}".format(
                    SyncStatus.describe(status), status, occurrences
                )
                self.console.new_line(
                    message, fg_color=self.status_fg_color(status)
                )
        else:
            self.console.new_line(u"\tNo changes!")

    def error(self, message):
        """Stops processing by raising an exception"""
        raise self.error_class(message)

    def confirm(self, *args, **kwargs):
        """Asks user for confirmation, providing boolean output"""
        if self.no_interactive:
            choice = True
        else:
            # Delimiter:
            self.console.new_line(
                u"{s} CONFIRMATION REQUIRED: {s}".format(s='#' * 28)
            )
            # Show diff if available:
            diff = kwargs.get('diff')
            if diff:
                self.console.new_line(u"{s} DIFF {s}".format(s='-' * 37))
                for diff_line in diff.splitlines():
                    if diff_line.startswith('+'):
                        fg_color = 'GREEN'
                    elif diff_line.startswith('-'):
                        fg_color = 'RED'
                    else:
                        fg_color = None
                    self.console.new_line(diff_line, fg_color=fg_color)
                self.console.new_line(u"{s}".format(s='-' * 80))
            # Show other messages:
            message = kwargs.get('message')
            messages = [message] if message else list(args)
            messages.append(
                kwargs.get(
                    'question', u"Do you want to apply this change?"
                )
            )
            for message in messages:
                self.console.new_line(message)
            choice = None
            while choice is None:
                user_text = self.get_input(u"Y/N? ")
                if user_text in ('Y', 'y'):
                    choice = True
                elif user_text in ('N', 'n'):
                    choice = False
            # Delimiter:
            self.console.new_line(u"{s}".format(s='#' * 80))
        return choice

    def run(self):
        """Performs the operation"""
        raise NotImplementedError


class WebsiteDumpOperation(BaseSyncOperation):
    """DUMPS website Pages TO structure of directories and file"""

    def dump_existing_pages(self, root_page, summary):
        """Dumps content of root page and children"""
        valid_paths = set([app_settings.SYNC_DIRECTORY])
        page_iterator = \
            itertools.chain([root_page], root_page.descendants().iterator())
        # Dump content of root page and children:
        for page in page_iterator:
            page_dumper = PageFileDumper(page)
            status = page_dumper.status()
            if status != SyncStatus.NO_CHANGES:
                if status == SyncStatus.ADDED:
                    apply_change = self.confirm(
                        u"Page created: {0}".format(page.url)
                    )
                else:
                    apply_change = self.confirm(
                        u"Page modified: {0}".format(page.url),
                        diff=page_dumper.diff()
                    )
            else:
                apply_change = True
            if not apply_change:
                status += SyncStatus.SKIPPED
            self.log_status(status, page_dumper.relative_path())
            if apply_change and not self.dry_run:
                page_dumper.save()
            # Update the set of valid paths (not to be deleted)
            valid_path = page_dumper.absolute_path()
            while valid_path != app_settings.SYNC_DIRECTORY:
                valid_paths.add(valid_path)
                valid_path = os.path.dirname(valid_path)
            summary[status] += 1
        return valid_paths

    def delete_unused_files(self, root_page, valid_paths, summary):
        """Deletes unused files ONLY in subtree of root"""
        if root_page.children():
            delete_start_path = os.path.dirname(
                PageFileDumper(root_page).absolute_path()
            )
            delete_list = []
            for root, dirs, files in os.walk(delete_start_path, False):
                for file_name in reversed(files):
                    if file_name.startswith('.'):
                        continue
                    path = os.path.join(root, file_name)
                    if path not in valid_paths:
                        delete_list.append(path)
                if root not in valid_paths:
                    delete_list.append(root)
            for delete_path in delete_list:
                relative_path = \
                    os.path.relpath(delete_path, app_settings.SYNC_DIRECTORY)
                status = SyncStatus.DELETED
                apply_change = self.confirm(
                    u"Path to be deleted: {0}".format(relative_path)
                )
                if not apply_change:
                    status += SyncStatus.SKIPPED
                self.log_status(status, relative_path)
                if apply_change and not self.dry_run:
                    if os.path.isdir(delete_path):
                        os.rmdir(delete_path)
                    else:
                        os.remove(delete_path)
                summary[status] += 1

    def add_to_vcs(self, summary):
        if (
            self.git_add and
            (SyncStatus.DELETED in summary or SyncStatus.ADDED in summary) and
            not self.dry_run and
            self.confirm(
                question=(
                    u"Do you want to add created and removed files to GIT?"
                )
            )
        ):
            output, errors = subprocess.Popen(
                ['git', '-C', app_settings.SYNC_DIRECTORY,
                 'add', '-A', app_settings.SYNC_DIRECTORY],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ).communicate()
            if errors:
                raise self.error(u"Adding file changes to GIT failed!")

    def run(self):
        """Performs the operation"""
        root_page = Page.objects.filter(url=self.root_url).first()
        if not root_page:
            self.error(u'Root page "{0}" not found!'.format(self.root_url))
        summary = collections.defaultdict(int)
        valid_paths = self.dump_existing_pages(root_page, summary)
        self.delete_unused_files(root_page, valid_paths, summary)
        self.summary(summary)
        self.add_to_vcs(summary)


class WebsiteLoadOperation(BaseSyncOperation):
    """LOADS website Pages FROM structure of directories and file"""

    def page_loaders(self):
        """Builds a list of page loaders"""
        root_path = url_to_path(self.root_url, has_children=False)
        if (
            os.path.exists(
                os.path.join(app_settings.SYNC_DIRECTORY, root_path)
            ) and
            self.root_url != '/'
        ):
            page_loaders = [FilePageLoader(root_path)]
        else:
            root_path = url_to_path(self.root_url, has_children=True)
            if not os.path.exists(
                os.path.join(app_settings.SYNC_DIRECTORY, root_path)
            ):
                self.error(u'Root page "{0}" not found!'.format(self.root_url))
            page_loaders = []
            walk_start = os.path.join(
                app_settings.SYNC_DIRECTORY, os.path.dirname(root_path)
            )
            if walk_start.endswith('/'):
                walk_start = walk_start[:-1]
            for root, dirs, files in os.walk(walk_start):
                for file_name in files:
                    if file_name.startswith('.'):
                        continue
                    absolute_path = os.path.join(root, file_name)
                    relative_path = os.path.relpath(
                        absolute_path, app_settings.SYNC_DIRECTORY
                    )
                    page_loaders.append(FilePageLoader(relative_path))
        return page_loaders

    def load_existing_files(self, summary):
        """Loads pages from existing files"""
        valid_pks = set()
        page_loaders = self.page_loaders()
        for page_loader in page_loaders:
            status = page_loader.status()
            if status != SyncStatus.NO_CHANGES:
                confirm_msgs, confirm_opts = [], {}
                if status == SyncStatus.ADDED:
                    page_is_dirty = False
                    confirm_msgs.append(
                        u"Page created: {0}".format(page_loader.url())
                    )
                else:
                    page_is_dirty = page_loader.page().is_dirty
                    confirm_msgs.append(
                        u"Page modified: {0}".format(page_loader.url())
                    )
                    confirm_opts['diff'] = page_loader.diff()
                    if page_is_dirty:
                        confirm_msgs.append(u"WARNING: Modified in Admin!")
                if page_is_dirty and not self.force:
                    apply_change = False
                else:
                    apply_change = self.confirm(*confirm_msgs, **confirm_opts)
            else:
                page_is_dirty = False
                apply_change = True
            if not apply_change:
                status += SyncStatus.SKIPPED
            elif page_is_dirty:
                status += SyncStatus.FORCED
            self.log_status(status, page_loader.url())
            if apply_change and not self.dry_run:
                page_loader.save()
            # Update the set of valid PKs (not to be deleted)
            loaded_page = page_loader.page()
            if loaded_page:
                valid_pks.add(loaded_page.pk)
            summary[status] += 1
        return valid_pks

    def delete_unused_pages(self, valid_pks, summary):
        """Removes unused pages only in subtree of root"""
        pages_to_delete = Page.objects.filter(
            url__startswith=self.root_url
        ).exclude(
            pk__in=valid_pks
        )
        for page in pages_to_delete:
            status = SyncStatus.DELETED
            confirm_msgs = [
                u"Page to be deleted: {0} ({1})".format(page.url, page.pk)
            ]
            if page.is_dirty:
                confirm_msgs.append(u"WARNING: Modified in Admin!")
            if page.is_dirty and not self.force:
                apply_change = False
            else:
                apply_change = self.confirm(*confirm_msgs)
            if not apply_change:
                status += SyncStatus.SKIPPED
            elif page.is_dirty:
                status += SyncStatus.FORCED
            self.log_status(status, page.url)
            if apply_change and not self.dry_run:
                page.delete()
            summary[status] += 1

    def run(self):
        """Performs the operation"""
        summary = collections.defaultdict(int)
        valid_pks = self.load_existing_files(summary)
        self.delete_unused_pages(valid_pks, summary)
        if self.dry_run:
            self.log(
                u"WARNING: Number of deleted records may be inadequate "
                u"when --dry-run is used!"
            )
        self.summary(summary)



# -*- coding: utf-8 -*-

import re

from django.test import TestCase
from django.test.utils import override_settings
from django.core.urlresolvers import reverse

from powerpages.models import Page


# TODO: tests for sitemap config options


def parse_sitemap(content):
    urlset_match = re.search(
        r'<urlset[^>]*>(?P<urls>[\s\S]*)</urlset>', content
    )
    if urlset_match:
        results = []
        urlset_content = urlset_match.groupdict()['urls']
        for url_content in re.findall(r'<url>([\s\S]+)</url>', urlset_content):
            results.append(
                dict(
                    re.findall(r'<([^>]+)>([^<]*)</[^>]+>', url_content)
                )
            )
    else:
        results = None
    return results


class PageSitemapTestCase(TestCase):

    def test_sitemap_view_empty(self):
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(parse_sitemap(response.content), [])

    def test_sitemap_defaults(self):
        page = Page.objects.create(
            url='/test/',
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': page.changed_at.strftime('%Y-%m-%d')
                }
            ]
        )

    def test_sitemap_page_excluded(self):
        Page.objects.create(
            url='/test/',
            page_processor_config={
                'sitemap': False
            }
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(parse_sitemap(response.content), [])

    def test_sitemap_lastmod_per_page(self):
        page = Page.objects.create(
            url='/test/',
            page_processor_config={
                'sitemap': {
                    'lastmod': "2012-12-30",
                }
            }
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': "2012-12-30",
                }
            ]
        )

    def test_sitemap_changefreq_per_page(self):
        page = Page.objects.create(
            url='/test/',
            page_processor_config={
                'sitemap': {
                    'changefreq': "monthly",
                }
            }
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': page.changed_at.strftime('%Y-%m-%d'),
                    'changefreq': "monthly"
                }
            ]
        )

    @override_settings(POWER_PAGES={'SITEMAP_DEFAULT_CHANGEFREQ': 'daily'})
    def test_sitemap_changefreq_default(self):
        page = Page.objects.create(
            url='/test/',
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': page.changed_at.strftime('%Y-%m-%d'),
                    'changefreq': "daily"
                }
            ]
        )

    @override_settings(POWER_PAGES={'SITEMAP_DEFAULT_CHANGEFREQ': 'daily'})
    def test_sitemap_changefreq_per_page_with_default(self):
        page = Page.objects.create(
            url='/test/',
            page_processor_config={
                'sitemap': {
                    'changefreq': "monthly",
                }
            }
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': page.changed_at.strftime('%Y-%m-%d'),
                    'changefreq': "monthly"  # page config has higher priority
                }
            ]
        )

    def test_sitemap_priority_per_page(self):
        page = Page.objects.create(
            url='/test/',
            page_processor_config={
                'sitemap': {
                    'priority': 0.5,
                }
            }
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': page.changed_at.strftime('%Y-%m-%d'),
                    'priority': "0.5"
                }
            ]
        )

    @override_settings(POWER_PAGES={'SITEMAP_DEFAULT_PRIORITY': 0.8})
    def test_sitemap_priority_default(self):
        page = Page.objects.create(
            url='/test/',
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': page.changed_at.strftime('%Y-%m-%d'),
                    'priority': "0.8"
                }
            ]
        )

    @override_settings(POWER_PAGES={'SITEMAP_DEFAULT_PRIORITY': 0.8})
    def test_sitemap_priority_per_page_with_default(self):
        page = Page.objects.create(
            url='/test/',
            page_processor_config={
                'sitemap': {
                    'priority': 0.5,
                }
            }
        )
        url = reverse('sitemap')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            parse_sitemap(response.content),
            [
                {
                    'loc': 'http://testserver{0}'.format(page.url),
                    'lastmod': page.changed_at.strftime('%Y-%m-%d'),
                    'priority': "0.5",  # page config has higher priority
                }
            ]
        )



# -*- coding: utf-8 -*-

from powerpages.utils.class_registry.formfields import (
    BaseRegistryItemChoiceField
)
from powerpages import page_processor_registry


class PageProcessorChoiceField(BaseRegistryItemChoiceField):
    """
    Form field intended to display keys from page_processor_registry.registry
    as it's choices
    """
    registry_instance = page_processor_registry.registry



# -*- coding: utf-8 -*-

from django.test import TestCase
from django.contrib.auth.models import User

from powerpages.models import Page
from powerpages.sync import PageFileDumper
from powerpages.admin import website_link, sync_status, save_page
from powerpages.signals import page_edited
from .test_sync import BaseSyncTestCase


class WebsiteLinkTestCase(TestCase):

    def test_no_object(self):
        self.assertIsNone(website_link(None))

    def test_empty_url(self):
        self.assertEqual(
            website_link(Page(url='')),
            u'<a href="" style="font-weight: normal;"> &raquo;</a>'
        )

    def test_root_url(self):
        self.assertEqual(
            website_link(Page(url='/')),
            u'<a href="/" style="font-weight: normal;">/ &raquo;</a>'
        )

    def test_first_level_url(self):
        self.assertEqual(
            website_link(Page(url='/test/')),
            u'<a href="/test/" style="font-weight: normal;">'
            u'/<span style="font-weight: bold">test</span>/'
            u' &raquo;</a>'
        )

    def test_second_level_url(self):
        self.assertEqual(
            website_link(Page(url='/nested/test/')),
            u'<a href="/nested/test/" style="font-weight: normal;">'
            u'/nested/<span style="font-weight: bold">test</span>/'
            u' &raquo;</a>'
        )

    def test_file(self):
        self.assertEqual(
            website_link(Page(url='/robots.txt')),
            u'<a href="/robots.txt" style="font-weight: normal;">'
            u'/<span style="font-weight: bold">robots.txt</span>'
            u' &raquo;</a>'
        )

    def test_nested_file(self):
        self.assertEqual(
            website_link(Page(url='/nested/robots.txt')),
            u'<a href="/nested/robots.txt" style="font-weight: normal;">'
            u'/nested/<span style="font-weight: bold">robots.txt</span>'
            u' &raquo;</a>'
        )


class SyncStatusTestCase(BaseSyncTestCase):

    def test_no_object(self):
        self.assertIsNone(sync_status(None))

    def test_file_synced(self):
        page = Page.objects.create(
            url='/test-page/', template='<h1>Test Page</h1>'
        )
        PageFileDumper(page).save()
        self.assertEqual(
            sync_status(page),
            u'<span style="color: green">File is synced</span>'
        )

    def test_file_content_differs(self):
        page = Page.objects.create(
            url='/test-page/', template='<h1>Test Page</h1>'
        )
        PageFileDumper(page).save()
        page.title = u"Lorem Ipsum"
        page.save()
        self.assertEqual(
            sync_status(page),
            u'<span style="color: orange">File content differs</span>'
        )

    def test_file_is_missing(self):
        page = Page.objects.create(
            url='/test-page/', template='<h1>Test Page</h1>'
        )
        self.assertEqual(
            sync_status(page),
            u'<span style="color: red">File is missing</span>'
        )

    def test_file_content_differs_modified_in_admin(self):
        page = Page.objects.create(
            url='/test-page/', template='<h1>Test Page</h1>'
        )
        PageFileDumper(page).save()
        page.title = u"Lorem Ipsum"
        page.is_dirty = True  # modified in Admin
        page.save()
        self.assertEqual(
            sync_status(page),
            u'<span style="color:black; font-weight:bold">'
            u'Changed in Admin!</span><br>'
            u'<span style="color: orange">File content differs</span>'
        )


class SavePageTestCase(TestCase):

    def setUp(self):

        def page_edited_test_handler(sender, **kwargs):
            self.page_edited_kwargs = kwargs

        self.page_edited_kwargs = None
        page_edited.connect(
            page_edited_test_handler, dispatch_uid='test_page_edited',
            weak=False
        )

    def tearDown(self):
        page_edited.disconnect(dispatch_uid='test_page_edited')
        self.page_edited_kwargs = None

    def test_create_page(self):
        page = Page(url='/test-page/')
        user = User.objects.create_user('admin-user')

        save_page(page=page, user=user, created=True)

        self.assertIsNotNone(page.pk)
        self.assertTrue(page.is_dirty)
        self.assertDictContainsSubset(
            {'page': page, 'user': user, 'created': True},
            self.page_edited_kwargs
        )

    def test_modify_page(self):
        page = Page.objects.create(url='/test-page/', title='Lorem')
        page.title = 'Ipsum'
        user = User.objects.create_user('admin-user')

        save_page(page=page, user=user, created=False)

        self.assertEqual(Page.objects.get(pk=page.pk).title, 'Ipsum')
        self.assertTrue(page.is_dirty)
        self.assertDictContainsSubset(
            {'page': page, 'user': user, 'created': False},
            self.page_edited_kwargs
        )



# -*- coding: utf-8 -*-

from django import http
from django.template import RequestContext
from django.template.loader import render_to_string
from django.contrib.auth.decorators import user_passes_test
from django.core.cache import cache

from powerpages.models import Page
from powerpages import sitemap_config
from powerpages import cachekeys


def page(request, path):
    """Page processing view"""
    # ensure that path starts and ends with "/"
    if not path.startswith("/"):
        path = "/" + path
    # redirect to equivalent page with ending slash
    # if path doesn't end with slash and it's not a file name:
    if not path.endswith("/") and '.' not in path.split('/')[-1]:
        return http.HttpResponsePermanentRedirect(path + "/")
    matching_pages = Page.objects.all().filter(url=path)
    try:
        page_obj = matching_pages[0]
    except IndexError:
        raise http.Http404
    page_processor = page_obj.get_page_processor()
    return page_processor.process_request(request)


@user_passes_test(lambda u: u.is_staff or u.is_superuser)
def admin_switch_edit_mode(request):
    """Turns website edit mode ON / OFF"""
    key = 'WEBSITE_EDIT_MODE'
    if key in request.session:
        del request.session[key]
    else:
        request.session[key] = True
    referrer = request.META.get('HTTP_REFERER')
    if referrer == request.path_info:
        referrer = None
    return http.HttpResponseRedirect(referrer or '/')


def sitemap(request):
    """Simple XML sitemap view."""
    sitemap_content = cache.get(cachekeys.SITEMAP_CONTENT)
    if sitemap_content is None:
        sitemap_content = render_to_string(
            'powerpages/sitemap.xml',
            {'urlset': sitemap_config.sitemaps.urls(request)},
            context_instance=RequestContext(request)
        )
        cache.get(cachekeys.SITEMAP_CONTENT, sitemap_content)
    return http.HttpResponse(sitemap_content, content_type='application/xml')



# -*- coding: utf-8 -*-

import datetime

from powerpages import sitemap_config
from powerpages.models import Page

VALID_CHANGEFREQ = (
    '', 'always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never',
)


class PageSitemap(sitemap_config.Sitemap):
    """Sitemap configuration for powerpages.Page model"""

    def get_items(self):
        for page in Page.objects.all():
            if not page.is_accessible():
                continue  # No sitemap for inaccessible Pages
            page_processor = page.get_page_processor()
            sitemap_settings = page_processor.config.get('sitemap')
            conf_item = {}
            if isinstance(sitemap_settings, dict):
                if 'lastmod' in sitemap_settings:
                    try:
                        lastmod = datetime.datetime.strptime(
                            unicode(sitemap_settings['lastmod']), '%Y-%m-%d'
                        )
                    except ValueError:
                        pass
                    else:
                        conf_item['lastmod'] = lastmod
                if 'priority' in sitemap_settings:
                    try:
                        priority = float(sitemap_settings['priority'])
                    except (ValueError, TypeError):
                        pass
                    else:
                        conf_item['priority'] = priority
                if 'changefreq' in sitemap_settings:
                    changefreq = sitemap_settings['changefreq']
                    if changefreq in VALID_CHANGEFREQ:
                        conf_item['changefreq'] = changefreq
            if sitemap_settings is not None and not sitemap_settings:
                # sitemap settings defined, but false in boolean context
                continue
            conf_item['location'] = page.url
            if 'lastmod' not in conf_item:
                conf_item['lastmod'] = page.changed_at
            yield conf_item


sitemap_config.sitemaps.add(PageSitemap)



# -*- coding: utf-8 -*-

from optparse import make_option

from django.core.management.base import BaseCommand, CommandError


class BaseDumpLoadCommand(BaseCommand):
    """Base class for website dump / load commands"""

    option_list = BaseCommand.option_list + (
        make_option(
            '--dry-run',
            action='store_true',
            default=False,
            dest='dry_run',
            help=(
                "Performs all operations without changing anything in the "
                "file system / database."
            )
        ),
        make_option(
            '--no-interactive',
            action='store_true',
            default=False,
            dest='no_interactive',
            help=(
                "Do not ask user to confirm changes in file system / database."
            )
        ),
        make_option(
            '--quiet',
            action='store_true',
            default=False,
            dest='quiet',
            help=(
                "Limits displayed output to summary of changes"
            )
        ),
        make_option(
            '-f', '--force',
            action='store_true',
            default=False,
            dest='force',
            help="Allows to overwite changes made in Admin."
        ),
        make_option(
            '--git-add',
            action='store_true',
            default=False,
            dest='git_add',
            help="Add created / removed files to GIT."
        )
    )

    operation_class = None  # must be set in subclass

    def handle(self, root_url='/', **options):
        """Performs the operation"""
        operation = self.operation_class(
            root_url=root_url, error_class=CommandError,
            stdout=self.stdout, stderr=self.stderr, **options
        )
        return operation.run()
